## Claus Hunsen, 2016
## hunsen@fim.uni-passau.de

OUTPUT = "png" # one of: pdf, png

## / / / / / / / / / / / / / /
## GLOBAL R CONFIGURATION
##

# disable stringsAsFactors
options(stringsAsFactors = FALSE)
# reset warnings
assign("last.warning", NULL, envir = baseenv())


## / / / / / / / / / / / / / /
## Set environment for script
##

## source script files
source(file.path("scripts/util-init.R"), chdir = TRUE) # source codeface-extraction-r

source(file.path("analysis-cli.R")) # command-line parameters
source(file.path("util.R")) # get our utility functions
source(file.path("logger.R")) # logging


## / / / / / / / / / / / / / /
## CLI AND DEBUGGING ---------
##

DEBUGGING = Sys.getenv("RSTUDIO") == 1 # set to TRUE for debugging in RStudio
if (DEBUGGING) {

    ## set command line arguments here
    cli.args = list(
        # "script" = "data", # data, null, sensitivity, nullcheck, stats, history-plots
        "script" = "sensitivity",
        # "script" = "history-plots",

        "casestudy" = "busybox",
        "artifact" = "feature",
        "artifact_relation" = "cochange",
        "selection_process" = "threemonth",
        "split_window" = "3 months",
        "sliding_window" = FALSE,
        "null_model" = "rewire", # one of rewire, random, errg
        "no_remove_core" = !FALSE, # do I want core developers? if yes, !TRUE; if no, !FALSE ;)

        "complete_rerun" = FALSE,

        "codeface_data" = "/scratch/codeface/codeface-data",
        #"codeface_data" = "/local/hunsen/projects/codeface-data",

        "loglevel" = "DEBUG"
    )

    if (cli.args[["script"]] %in% c("null", "sensitivity")) {
        Sys.setenv("SLURM_ARRAY_TASK_ID" = 24)
    } else {
        Sys.unsetenv("SLURM_ARRAY_TASK_ID")
    }

} else {

    ## command-line arguments
    cli.parser = create.cli.parser()
    cli.args = cli.parser$parse_args()

}

## COMMAND-LINE ARGUMENTS
SCRIPT = cli.args[["script"]]
CASESTUDY = cli.args[["casestudy"]]
AUTHOR.RELATION = "mail"
ARTIFACT = cli.args[["artifact"]]
ARTIFACT.RELATION = cli.args[["artifact_relation"]]
SELECTION.PROCESS = cli.args[["selection_process"]]
SPLIT.WINDOW = cli.args[["split_window"]]
CODEFACE.DATA = cli.args[["codeface_data"]]
RERUN.ANALYSIS = cli.args[["complete_rerun"]]
SLIDING.WINDOW = cli.args[["sliding_window"]]
NULL.MODEL = cli.args[["null_model"]]
REMOVE.CORE = !cli.args[["no_remove_core"]]
LOG.LEVEL = cli.args[["loglevel"]]
CALLGRAPHS = ARTIFACT.RELATION == "callgraph"


## / / / / / / / / / / / / / /
## OUTPUT PATHS --------------
##

SPLIT.WINDOW.STRING = switch(
    SELECTION.PROCESS,
    threemonth = gsub(" ", "_", SPLIT.WINDOW),
    releases = "releases",
    SELECTION.PROCESS
)

TMP = file.path(getwd(), "data-v3", SPLIT.WINDOW.STRING)


## / / / / / / / / / / / / / /
## DATA PREPARATION ----------
##

## global data structures
results = list()

## construct analysis identifier
ANALYSIS = construct.analysis.id(CASESTUDY, ARTIFACT, ARTIFACT.RELATION)
ANALYSIS.NUMBER = Sys.getenv("SLURM_ARRAY_TASK_ID", unset = "")

## create temporary path for storage of intermediate data
TMP.CASESTUDY = file.path(TMP, CASESTUDY)
TMP.ANALYSIS = file.path(TMP.CASESTUDY, ANALYSIS)
dir.create(TMP.ANALYSIS, recursive = TRUE, showWarnings = FALSE)
TMP.RANGES =  file.path(TMP.CASESTUDY, "ranges")
dir.create(TMP.RANGES, recursive = TRUE, showWarnings = FALSE)
## create null model folders
NULL.STRING = sprintf("null-%s", NULL.MODEL)
TMP.NULL = file.path(TMP.ANALYSIS, NULL.STRING)
if (REMOVE.CORE) {
    TMP.NULL = paste0(TMP.NULL, "-without-core")
}
dir.create(TMP.NULL, recursive = TRUE, showWarnings = FALSE)
TMP.NULL.ORIGINAL = file.path(TMP.NULL, "original")
dir.create(TMP.NULL.ORIGINAL, recursive = TRUE, showWarnings = FALSE)
## Sensitivity analysis per Kossinets
TMP.SENSITIVITY = file.path(TMP.ANALYSIS, "sensitivity")
dir.create(TMP.SENSITIVITY, recursive = TRUE, showWarnings = FALSE)
## create path for plots and results
TMP.RESULTS = file.path(TMP, "results")
dir.create(TMP.RESULTS, recursive = TRUE, showWarnings = FALSE)
TMP.PLOT = file.path(TMP.RESULTS, "history-plots")
dir.create(TMP.RESULTS, recursive = TRUE, showWarnings = FALSE)
TMP.STATS = file.path(TMP.RESULTS, "stats")
dir.create(TMP.STATS, recursive = TRUE, showWarnings = FALSE)

## path for configuration
PATH.DATA.CONF = file.path(TMP.ANALYSIS, paste(ANALYSIS, "conf", sep = "-"))


## / / / / / / / / / / / / / /
## LOGGING -------------------
##

log.folder = file.path("logs-v3", SPLIT.WINDOW.STRING, CASESTUDY, paste0(ANALYSIS, "_", SCRIPT))
log.null.model = if (SCRIPT == "null") paste0("-", NULL.MODEL) else ""
log.number = if (ANALYSIS.NUMBER != "") paste0("_", ANALYSIS.NUMBER) else ""

if (DEBUGGING) {
    log.file = file.path(log.folder, paste0(ANALYSIS, "_", SCRIPT, log.null.model, log.number, ".slurm"))
} else {
    log.file = NULL
}

log.create(log.level = LOG.LEVEL, log.file = log.file, log.file.remove = TRUE, log.console = TRUE)
logging::loginfo("Initiated logger with log level %s", LOG.LEVEL)


## / / / / / / / / / / / / / /
## PARALLELIZATION -----------
##

library(parallel)

if (DEBUGGING) {
    options(mc.cores = 1) # 1 core leads to the use of lapply (instead of mclapply)
    logging::loginfo("Configuring no parallelization.")
} else {
    cores = max(4L, detectCores() - 4, na.rm = TRUE)
    options(mc.cores = cores) # use this number of cores for mclapply
    logging::loginfo("Configuring parallelization on %s cores.", cores)
}


## / / / / / / / / / / / / / /
## CONFIGURATION -------------
##

source("configuration.R") # network configuration


## / / / / / / / / / / / / / /
## DISPATCH SUBSCRIPTS -------
##

## identify script
SCRIPT = match.arg(SCRIPT, c("data", "null", "sensitivity", "nullcheck", "stats", "history-plots"))

## dispatch!
if (SCRIPT == "data") {
    source("data.R")

    ## CONFIGURATION
    conf = ProjectConf$new(CODEFACE.DATA, SELECTION.PROCESS, CASESTUDY, ARTIFACT)
    conf$update.values(PROJCONF.VALUES)

    ## RANGE-LEVEL DATA
    PATH.DATA.SPARSE = file.path(TMP.ANALYSIS, paste(ANALYSIS, "data.sparse", sep = "-"))
    data = load.and.save("data", PATH.DATA.SPARSE, skip = RERUN.ANALYSIS, if.not.found = function() {
        if (SELECTION.PROCESS == "threemonth") {
            cf.data = ProjectData$new(conf)
            data = split.data.time.based(cf.data, SPLIT.WINDOW, split.basis = "commits", sliding.window = SLIDING.WINDOW)
        } else if(SELECTION.PROCESS == "releases") {
            data = construct.data(conf, callgraphs = CALLGRAPHS)
        }
        ## save updated configuration
        save(conf, file = PATH.DATA.CONF)
        return(data)
    })

    ## analysis
    motif.data = get.motif.data(data)
    save.data(conf, data, motif.data)

    # PROJECT-LEVEL DATA
    # PATH.DATA.GLOBAL.SPARSE = file.path(TMP.ANALYSIS, paste(ANALYSIS, "data.global.sparse", sep = "-"))
    # data.global = list("global" =
    #                        load.and.save("data.global", PATH.DATA.GLOBAL.SPARSE, skip = RERUN.ANALYSIS, if.not.found = function() {
    #                            data = ProjectData$new(conf)
    #                            attr(data, "range") = "global"
    #                            return(data)
    #                        })
    #                    )
    #
    # motif.data.global = get.motif.data(data.global)
    # save.data(conf, data.global, motif.data.global, venn = TRUE)

} else if (SCRIPT == "null") {
    source("null.R")

    if (ANALYSIS.NUMBER == "") {
        logging::logerror("Slurm array task ID could not be determined.")
        stop("Aborting...")
    } else if (as.integer(ANALYSIS.NUMBER) >= NULL.ITERATIONS) {
        logging::logerror("Number of needed iterations (%s) already reached (%s given).", NULL.ITERATIONS, ANALYSIS.NUMBER)
        stop("Aborting...")
    }

    ## load configuration
    load(file = PATH.DATA.CONF)

    ## calculate original model (only when removing core devs)
    if (REMOVE.CORE) {
        logging::loginfo("Running original-model.")
        calculate.null.model.original.for.range.data(conf)
    }

    ## calculate null model
    logging::loginfo("Running null-model iteration %s.", ANALYSIS.NUMBER)
    motif.counts = calculate.null.model(conf, NETCONF, ANALYSIS.NUMBER)

    save.data(conf, NULL, motif.counts, base.path = TMP.NULL, analysis.suffix = ANALYSIS.NUMBER)

    ## combine results for all null models
    if (ANALYSIS.NUMBER == NULL.ITERATIONS - 1) {

        ## collect all result files (ANALYSIS-\d{2})
        PATH.NULL.RESULTS = paste(ANALYSIS, "\\d(\\d?).csv", sep = "-")
        files = list.files(path = TMP.NULL, pattern = PATH.NULL.RESULTS)
        files = file.path(TMP.NULL, files)

        ## combine data by calling rbind.fill

        results = lapply(files, read.frequencies.from.disk)
        results = plyr::rbind.fill(results)
        results[is.na(results)] = 0

        ## aggregate data by calculating mean value over all iterations
        results.agg = aggregate(subset(results, select =-c(range)), list(range = results[["range"]]), mean, na.rm = TRUE)

        ## re-calculate f-metric
        for (motif in MOTIFS) {
            results.agg[paste0(motif, ".p1")] = results.agg[paste0(motif, ".mailing")] / results.agg[paste0(motif, ".complete")]
            results.agg[paste0(motif, ".p2")] = results.agg[paste0(motif, ".collaborating.and.mailing")] /
                results.agg[paste0(motif, ".collaborating")]
        }

        ## dump aggregated null-model data to disk
        write.csv(results.agg, file = file.path(TMP.NULL, paste0(ANALYSIS, ".csv")), fileEncoding = "utf-8", row.names = FALSE)
    }

} else if (SCRIPT == "sensitivity") {
    source("sensitivity.R")

    if (ANALYSIS.NUMBER == "") {
        logging::logerror("Slurm array task ID could not be determined.")
        stop("Aborting...")
    } else if (as.integer(ANALYSIS.NUMBER) >= NULL.ITERATIONS) {
        logging::logerror("Number of needed iterations (%s) already reached (%s given).", NULL.ITERATIONS, ANALYSIS.NUMBER)
        stop("Aborting...")
    }

    ## load configuration
    load(file = PATH.DATA.CONF)

    ## construct network models
    logging::loginfo("Construct random networks per Kossinets (iteration %s).", ANALYSIS.NUMBER)
    algo = "BSPC"
    construct.kossinets.networks(conf, ANALYSIS.NUMBER, algo)

    ## calculate Kossinets metrics
    metrics.kossinets = calculate.kossinets.model(conf, ANALYSIS.NUMBER, algo)
    save.kossinets.data(metrics.kossinets, algo, type = "kossinets", analysis.suffix = ANALYSIS.NUMBER)

    ## calculate motif metrics for Kossinets networks
    motifs.kossinets = calculate.motif.metrics.for.kossinets.models(conf, NETCONF, ANALYSIS.NUMBER, algo)
    save.kossinets.data(motifs.kossinets, algo, type = "motifs", analysis.suffix = ANALYSIS.NUMBER)

    ## combine results for all Kossinets models
    if (ANALYSIS.NUMBER == NULL.ITERATIONS - 1) {

        ## Kossinets metrics:
        ## 0) collect all result files (ANALYSIS-\d{2})
        PATH.KOSSINETS.RESULTS = paste(ANALYSIS, sprintf("kossinets-%s-\\d(\\d?).csv", algo), sep = "-")
        files = list.files(path = TMP.SENSITIVITY, pattern = PATH.KOSSINETS.RESULTS)
        files = file.path(TMP.SENSITIVITY, files)
        ## 1) combine data by calling rbind.fill
        metrics.kossinets.combined = lapply(files, read.frequencies.from.disk)
        metrics.kossinets.combined = plyr::rbind.fill(metrics.kossinets.combined)
        ## 2) aggregate data by calculating mean value over all iterations
        metrics.kossinets.combined.agg = aggregate(
            x = subset(metrics.kossinets.combined, select =-c(range, algo, step)),
            by = list(
                range = metrics.kossinets.combined[["range"]],
                step = metrics.kossinets.combined[["step"]],
                algo = metrics.kossinets.combined[["algo"]]
            ),
            FUN = mean,
            na.rm = TRUE
        )
        ## 3) dump aggregated Kossinets data to disk
        write.csv(
            metrics.kossinets.combined.agg,
            file = file.path(TMP.SENSITIVITY, paste0(ANALYSIS, "-kossinets", ".csv")),
            fileEncoding = "utf-8", row.names = FALSE
        )

        ## Motifs metrics:
        ## 0) collect all result files (ANALYSIS-\d{2})
        PATH.MOTIFS.RESULTS = paste(ANALYSIS, sprintf("motifs-%s-\\d(\\d?).csv", algo), sep = "-")
        files = list.files(path = TMP.SENSITIVITY, pattern = PATH.MOTIFS.RESULTS)
        files = file.path(TMP.SENSITIVITY, files)
        ## 1) combine data by calling rbind.fill
        metrics.motifs.combined = lapply(files, read.frequencies.from.disk)
        metrics.motifs.combined = plyr::rbind.fill(metrics.motifs.combined)
        ## 2) aggregate data by calculating mean value over all iterations
        metrics.motifs.combined.agg = aggregate(
            x = subset(metrics.motifs.combined, select =-c(range, algo, step)),
            by = list(
                range = metrics.motifs.combined[["range"]],
                step = metrics.motifs.combined[["step"]],
                algo = metrics.motifs.combined[["algo"]]
            ),
            FUN = mean,
            na.rm = TRUE
        )
        ## 3) recalculate p2 as it gets messed up during aggregation
        for (motif in MOTIFS) {
            metrics.motifs.combined.agg[paste(motif, "p2", sep = ".")] =
                metrics.motifs.combined.agg[[paste(motif, "collaborating.and.mailing", sep = ".")]] /
                metrics.motifs.combined.agg[[paste(motif, "collaborating", sep = ".")]]
        }
        ## 4) calculate relative errors
        relative.step.errors = plyr::ddply(metrics.motifs.combined.agg, .(range), function(df) {
            for (motif in MOTIFS) {
                ## default value
                df[paste(motif, "p2.step.error", sep = ".")] = NaN

                ## calculate relative error if possible
                original.value = df[df$step == 0, paste(motif, "p2", sep = ".")]
                if (!is.nan(original.value) && original.value != 0) {
                    df[paste(motif, "p2.step.error", sep = ".")] =
                        abs( df[[paste(motif, "p2", sep = ".")]] - original.value ) / original.value
                }
            }

            return(df[c("step", paste(MOTIFS, "p2.step.error", sep = "."))])
        })
        metrics.motifs.combined.agg = merge(metrics.motifs.combined.agg, relative.step.errors, sort = FALSE)
        ## 5) dump aggregated motifs data to disk
        write.csv(
            metrics.motifs.combined.agg,
            file = file.path(TMP.SENSITIVITY, paste0(ANALYSIS, "-motifs", ".csv")),
            fileEncoding = "utf-8", row.names = FALSE
        )
    }

} else if (SCRIPT == "nullcheck") {
    source("null.R")

    ## load configuration
    load(file = PATH.DATA.CONF)
    ## check null model
    check.null.model(conf)

} else if (SCRIPT == "stats") {
    source("stats.R")

    ## cache range selections
    for (CASESTUDY in CASESTUDIES) {
        logging::loginfo("Collecting filled and active ranges for %s.", CASESTUDY)

        TMP.RANGES.FILLED = file.path(TMP, CASESTUDY, "ranges", paste(CASESTUDY, "ranges.filled", sep = "-"))
        ranges.subset.filled = load.and.save("ranges.subset.filled", TMP.RANGES.FILLED, skip = RERUN.ANALYSIS, if.not.found = function() {
            ranges.subset.filled = get.filled.ranges(CASESTUDY)
            return(ranges.subset.filled)
        })
        TMP.RANGES.ACTIVE = file.path(TMP, CASESTUDY, "ranges", paste(CASESTUDY, "ranges.active", sep = "-"))
        ranges.subset.active = load.and.save("ranges.subset.active", TMP.RANGES.ACTIVE, skip = RERUN.ANALYSIS, if.not.found = function() {
            ranges.subset.active = get.active.ranges(CASESTUDY, threshold = "mean")
            return(ranges.subset.active)
        })
    }
    rm(ranges.subset.filled)
    rm(ranges.subset.active)

    ##
    ## PROJECT TABLE
    ##

    ## compose table for LaTeX
    TMP.STATS.TABLE.DATA = file.path(TMP.STATS, "table.data")
    tab = load.and.save("tab", TMP.STATS.TABLE.DATA, skip = RERUN.ANALYSIS, if.not.found = function() {
        tabs = lapply(CASESTUDIES, function(CASESTUDY) {
            ## get filled ranges
            TMP.RANGES.FILLED = file.path(TMP, CASESTUDY, "ranges", paste(CASESTUDY, "ranges.filled", sep = "-"))
            load(TMP.RANGES.FILLED)
            # ranges.subset.filled = get.filled.ranges(CASESTUDY)
            ## get table
            tab = get.table(CASESTUDY, ranges.subset = ranges.subset.filled, null.model = null.model)
            return(tab)
        })
        tab = rbind.fill(tabs)
        return(tab)
    })
    print.latex.table(tab)

    TMP.STATS.TABLE.DATA = file.path(TMP.STATS, "table.averages.developer.data")
    tab = load.and.save("tab", TMP.STATS.TABLE.DATA, skip = RERUN.ANALYSIS, if.not.found = function() {
        tabs = lapply(CASESTUDIES, function(CASESTUDY) {
            ## get table
            tab = get.table.averages.dev(CASESTUDY)
            return(tab)
        })
        tab = rbind.fill(tabs)
        return(tab)
    })
    print.latex.table.averages(tab, "developer")

    TMP.STATS.TABLE.DATA = file.path(TMP.STATS, "table.averages.artifact.data")
    tab = load.and.save("tab", TMP.STATS.TABLE.DATA, skip = RERUN.ANALYSIS, if.not.found = function() {
        tabs = lapply(CASESTUDIES, function(CASESTUDY) {
            ## get table
            tab = get.table.averages.artifacts(CASESTUDY)
            return(tab)
        })
        tab = rbind.fill(tabs)
        return(tab)
    })
    print.latex.table.averages(tab, "artifact")

    ##
    ## HYPOTHESIS 01
    ##

    ## get raw data for empirical data
    get.statistics.hypo1.collect()
    # get.statistics.hypo1.collect(subset = "active")

    ## get raw data for null data
    get.statistics.hypo1.collect(null.model = TRUE)
    # get.statistics.hypo1.collect(subset = "active", null.model = TRUE)

    ## only if iteration for null model
    get.statistics.hypo1()
    # get.statistics.hypo1(subset = "active")

    ##
    ## HYPOTHESIS 02
    ##

    ## all ranges
    get.statistics.hypo2()
    # get.statistics.hypo2(subset = "active")

    ##
    ## HYPOTHESIS 03
    ##

    get.statistics.hypo3()
    # get.statistics.hypo3(subset = "active")

    ##
    ## HYPOTHESIS 04
    ##

    get.statistics.hypo4()
    # get.statistics.hypo4(subset = "active")

} else if (SCRIPT == "history-plots") {
    source("history-plots.R")

    ## FIXME identify maximum number of filled ranges over all case studies and spread each to this number!

    ## load configuration
    load(file = PATH.DATA.CONF)

    ## cache range selections
    TMP.RANGES.FILLED = file.path(TMP.RANGES, paste(CASESTUDY, "ranges.filled", sep = "-"))
    ranges.subset.filled = load.and.save("ranges.subset.filled", TMP.RANGES.FILLED, skip = RERUN.ANALYSIS, if.not.found = function() {
        ranges.subset.filled = get.filled.ranges(CASESTUDY)
        return(ranges.subset.filled)
    })
    TMP.RANGES.ACTIVE = file.path(TMP.RANGES, paste(CASESTUDY, "ranges.active", sep = "-"))
    ranges.subset.active = load.and.save("ranges.subset.active", TMP.RANGES.ACTIVE, skip = RERUN.ANALYSIS, if.not.found = function() {
        ranges.subset.active = get.active.ranges(CASESTUDY, threshold = "mean")
        return(ranges.subset.active)
    })
    # ranges.subset.filled = get.filled.ranges(CASESTUDY)
    # ranges.subset.active = get.active.ranges(CASESTUDY, threshold = "mean")
    thresholds = attr(ranges.subset.active, "thresholds")
    ranges.subset.active = join.boolean.values(list(
        ranges.subset.filled, ranges.subset.active
    ), op = "and")

    ## plot
    # plot.venn.diagrams(conf, ranges.subset = ranges.subset.filled)
    # plot.evolution(ranges.subset = ranges.subset.filled, thresholds = NULL)
    # plot.evolution(ranges.subset = ranges.subset.active, bars = TRUE, thresholds = thresholds)
    # plot.null.models(conf, ranges.subset = ranges.subset.filled)
    plot.sensitivity.models(conf, ranges.subset = ranges.subset.filled, thresholds = NULL, smooth = TRUE)
}

cat("CR analysis done.", fill = TRUE)
