## Claus Hunsen, July 2015
## hunsen@fim.uni-passau.de


## / / / / / / / / / / / / / /
## Set environment for script
##

## libraries
library(plyr) # for rbind.fill.matrix
library(igraph) # graphs, baby!


remove.artifacts.from.matched.motifs = function(graph, vs) {
    vs.cleaned = lapply(vs, function(seq) {
        ## get types of nodes
        types = vertex_attr(graph, "type", index = seq)
        ## remove artifact nodes
        seq = seq[ types != TYPE.ARTIFACT ]
        return(seq)
    })
    return(vs.cleaned)
}

find.motifs.in.graph = function(graph, motif, color.attr = "type") {

    vs = subgraph_isomorphisms(target = graph, pattern = motif, method = "vf2",
                               vertex.color1 = vertex_attr(graph, color.attr), vertex.color2 = vertex_attr(motif, color.attr))

    ## normalize found vertex sequences
    vs.cleaned = lapply(vs, function(seq) {
        ## get types and names of nodes
        types = vertex_attr(graph, "type", index = seq)
        names = vertex_attr(graph, "name", index = seq)

        ## sort vertex sequence by types and names
        seq = seq[ order(types, names) ]

        return(seq)
    })

    ## identify duplicates in list of found vertex sequences
    vs.matrix = lapply(vs.cleaned, function(vs) {
        ## return the vertex sequence as matrix
        vs.matrix = matrix(vs, nrow = 1)
        return(vs.matrix)
    })
    vs.matrix = rbind.fill.matrix(vs.matrix)
    vs.duplicates = duplicated(vs.matrix)

    ## remove duplicates from list of vertex sequences
    vs.final = vs.cleaned[ !vs.duplicates ]

    return(vs.final)
}


count.motifs = function(net, motif.collaborating, motif.mailing, motif.collaborating.and.mailing) {

    ## get isomorphisms
    collaborating = find.motifs.in.graph(graph = net, motif = motif.collaborating, color.attr = "type")
    mailing = find.motifs.in.graph(graph = net, motif = motif.mailing, color.attr = "type")
    collaborating.and.mailing = find.motifs.in.graph(graph = net, motif = motif.collaborating.and.mailing, color.attr = "type")

    ## compute intersections for verification of matched sets
    cleaned.collaborating = unique(remove.artifacts.from.matched.motifs(net, collaborating))
    cleaned.collaborating.and.mailing = unique(remove.artifacts.from.matched.motifs(net, collaborating.and.mailing))
    correct = setequal(cleaned.collaborating.and.mailing, intersect(cleaned.collaborating, mailing))

    ## get list of developers
    devs.length = length(V(net)[ type == TYPE.AUTHOR ])
    artifacts.length = length(V(net)[ type == TYPE.ARTIFACT ])
    dev.pairs = choose(devs.length, 2)

    # return an object for later inspection
    return(list(
        "developers" = devs.length, # total number of developers
        "artifacts" = artifacts.length, # total number of artifacts
        "complete" = dev.pairs, # total number of developer pairs
        "collaborating" = length(collaborating), # number of developer pairs that collaborate
        "collaborating.raw" = collaborating,
        "mailing" = length(mailing), # number of developer pairs that exchange emails
        "mailing.raw" = mailing,
        "collaborating.and.mailing" = length(collaborating.and.mailing), # number of developer pairs that collaborate and exchange emails
        "collaborating.and.mailing.raw" = collaborating.and.mailing,
        "correct" = correct, # sanity check
        "p1" = length(mailing) / dev.pairs, # fraction p1
        "p2" = length(collaborating.and.mailing) / length(collaborating) # fraction p2
    ))

}


## / / / / / / / / / / / / / /
## Triangle motif
##

## positive triangle motif
triangle.positive = make_empty_graph(directed = FALSE) +
    vertices("D1", "D2", "A",
             type = c(TYPE.AUTHOR, TYPE.AUTHOR, TYPE.ARTIFACT)) +
    edges("D1","A" , "D2","A", "D1","D2",
          type = c(TYPE.EDGES.INTER, TYPE.EDGES.INTER, TYPE.EDGES.INTRA))

## negative triangle motif
triangle.negative = make_empty_graph(directed = FALSE) +
    vertices("D1", "D2", "A", type = c(TYPE.AUTHOR, TYPE.AUTHOR, TYPE.ARTIFACT)) +
    edges("D1","A" , "D2","A", type = c(TYPE.EDGES.INTER, TYPE.EDGES.INTER))

## developer relation
triangle.mailing = make_empty_graph(directed = FALSE) +
    vertices("D1", "D2", type = c(TYPE.AUTHOR, TYPE.AUTHOR)) +
    edges("D1","D2", type = c(TYPE.EDGES.INTRA))




## / / / / / / / / / / / / / /
## Square motif
##

## positive square motif
square.positive = make_empty_graph(directed = FALSE) +
    vertices("D1", "D2", "A1", "A2",
             type = c(TYPE.AUTHOR, TYPE.AUTHOR, TYPE.ARTIFACT, TYPE.ARTIFACT)) +
    edges("D1","A1" , "D2","A2", "A1","A2", "D1","D2",
          type = c(TYPE.EDGES.INTER, TYPE.EDGES.INTER, TYPE.EDGES.INTRA, TYPE.EDGES.INTRA))

## negative square motif
square.negative = make_empty_graph(directed = FALSE) +
    vertices("D1", "D2", "A1", "A2",
             type = c(TYPE.AUTHOR, TYPE.AUTHOR, TYPE.ARTIFACT, TYPE.ARTIFACT)) +
    edges("D1","A1" , "D2","A2", "A1","A2",
          type = c(TYPE.EDGES.INTER, TYPE.EDGES.INTER, TYPE.EDGES.INTRA))

## developer relation
square.mailing = make_empty_graph(directed = FALSE) +
    vertices("D1", "D2", type = c(TYPE.AUTHOR, TYPE.AUTHOR)) +
    edges("D1","D2", type = c(TYPE.EDGES.INTRA))


## / / / / / / / / / / / / / /
## Motif strategy
##

get.motif.counts = function(net) {

    # get motif counts
    triangle = count.motifs(net, triangle.negative, triangle.mailing, triangle.positive)
    square = count.motifs(net, square.negative, square.mailing, square.positive)

    # return raw data
    return(list(
        "triangle" = triangle,
        "square" = square
    ))

}
