## Claus Hunsen, July 2015
## hunsen@fim.uni-passau.de


## / / / / / / / / / / / / / /
## Set environment for script
##

## libraries
library(plyr) # for rbind.fill.matrix
library(igraph) # graphs, baby!
library(ggplot2) # plots
library(logging) # logging, obviously


## / / / / / / / / / / / / / /
## Input format
##

## return(list(
##     "triangle" = triangle,
##     "square" = square
## ))
##
## Each 'triangle' and 'square' looks like this:
##  return(list(
##      "developers" = devs.length, # total number of developers
##      "artifacts" = artifacts.length, # total number of artifacts
##      "complete" = dev.pairs, # total number of developer pairs
##      "collaborating" = length(collaborating), # number of developer pairs that collaborate
##      "collaborating.raw" = collaborating,
##      "mailing" = length(mailing), # number of developer pairs that exchange emails
##      "mailing.raw" = mailing,
##      "collaborating.and.mailing" = length(collaborating.and.mailing), # number of developer pairs that collaborate and exchange emails
##      "collaborating.and.mailing.raw" = collaborating.and.mailing,
##      "correct" = correct, # sanity check
##      "p1" = length(mailing) / length(dev.pairs), # fraction p1
##      "p2" = length(collaborating.and.mailing) / length(collaborating) # fraction p2
##  ))



## / / / / / / / / / / / / / /
## Transform motif data to data.frame
##

transform.motif.data.to.data.frame = function(motifs) {

    ## remove raw data and 'correct' flag
    to.remove = c("collaborating.and.mailing.raw", "mailing.raw", "collaborating.raw", "correct")

    # transform the inner list to a data.frame
    dat.list = lapply(motifs, function(motif) {
        stripped = motif[ !(names(motif) %in% to.remove) ]
        stripped.dat = as.data.frame(stripped)

        raw = motif[ names(motif) %in% to.remove ]
        names(raw) = gsub(".raw", "", names(raw), fixed = TRUE)
        attr(stripped.dat, "raw") = raw

        return(stripped.dat)
    })

    return(dat.list)

}

combine.data.to.single.data.frame = function(counts) {

    counts.df = lapply(counts, function(range.data) {

        # cbind all motifs into one data.frame
        dat.frame = as.data.frame(range.data)

    })
    # combine to single data.frame
    counts.df = rbind.fill(counts.df)

    # insert range data into data.frame as an independent row
    counts.df = cbind("range" = names(counts), counts.df)

    return(counts.df)
}


## / / / / / / / / / / / / / /
## Frequencies
##

get.frequency.table = function(artifact, artifact.relation, motifs, results, short = FALSE) {
    frequencies = list()

    for (motif in motifs) {
        df = combine.data.to.single.data.frame(results$motif.counts)

        if (short) {
            ## get subset of only a few ranges
            first = 1; mid = ceiling(nrow(df) / 2); last = nrow(df)
            df = df[ c(first, mid, last), ]
        }

        ## get ranges (use as column names)
        ranges = df[[ "range" ]]

        ## compute frequencies
        frequency = df[[ paste0(motif, ".collaborating.and.mailing") ]] / df[[ paste0(motif, ".collaborating") ]]
        freq = t(cbind(
            df[ paste0(motif, c(
                ".artifacts",
                ".developers",
                ".mailing",
                ".collaborating",
                ".collaborating.and.mailing",
                ".p1",
                ".p2"
                )) ],
            "frequency" = frequency
            ))

        ## adjust row and column names
        rownames(freq) = gsub(motif, artifact, rownames(freq), fixed = TRUE)
        rownames(freq) = gsub("collaborating.and.mailing", "coll+mail", rownames(freq), fixed = TRUE) # shorten that
        colnames(freq) = paste(ranges, artifact.relation, sep = ".")

        ## round frequencies
        freq = round(freq, digit = 3)

        ## append to global frequencies list
        frequencies[[ motif ]] = rbind.fill( frequencies[[ motif ]], freq )
    }

    return(frequencies)
}


write.frequencies.to.disk = function(casestudy, artifact.relation, motifs, frequencies, path, filepattern = "%s-%s-%s.csv") {
    ## write one file for each motif
    for (motif in motifs) {
        write.csv(t(frequencies[[ motif ]]), # extract current frequencies
                  file = file.path(path, sprintf(filepattern, casestudy, artifact.relation, motif)),
                  fileEncoding = "utf-8")
    }
}


read.frequencies.from.disk = function(file) {

    ## read one file for each motif
    frequencies = read.csv(file, fileEncoding = "utf-8")

    return(frequencies)
}


collect.frequencies.from.disk = function(casestudy, ranges.subset = NULL, cols = NULL, cols.rename = NULL) {

    results = list()
    for(artifact.relation in ARTIFACT.RELATIONS) {
        for (artifact in ARTIFACTS) {
            ## get temporary path for storage of intermediate data
            analysis = construct.analysis.id(casestudy, artifact, artifact.relation)
            tmp.folder = file.path(TMP, casestudy, analysis)
            data.file = file.path(tmp.folder, paste0(analysis, ".csv"))

            ## read data
            data = read.frequencies.from.disk(data.file)

            ## subset columns
            if (!is.null(cols)) {
                data = data[cols]
                if (!is.null(cols.rename))
                    colnames(data) = cols.rename
            }

            ## remove ranges
            if (!is.null(ranges.subset)) {
                data = data[ ranges.subset, ]
            }


            ## store in results
            results[[ casestudy ]][[ artifact.relation ]][[ artifact ]] = data
        }
    }

    return(results)
}

collect.frequencies.from.disk.metrics = function(casestudy, ranges.subset = NULL) {
    data = collect.frequencies.from.disk(
        casestudy,
        ranges.subset = ranges.subset,
        cols = c("range", "triangle.collaborating", "triangle.collaborating.and.mailing", "triangle.p2",
                 "square.collaborating", "square.collaborating.and.mailing", "square.p2"),
        cols.rename = c("range", "triangle.cr", "triangle.fcr", "triangle.f",
                        "square.cr", "square.fcr", "square.f")
    )
    return(data)
}


get.filled.ranges.for.conf = function(casestudy, artifact, artifact.relation) {
    analysis = construct.analysis.id(casestudy, artifact, artifact.relation)
    tmp.analysis = file.path(TMP, casestudy, analysis)

    ## path for configuration
    path.conf = file.path(tmp.analysis, paste(analysis, "conf", sep = "-"))
    ## load configuration
    load(file = path.conf)

    ## remove all ranges at start and end that do not have any e-mail activity or any commit activity
    start = 0; end = -1
    for (i in 1:length(conf$get.entry("ranges"))) {
        range = conf$get.entry("ranges")[i]
        PATH.DATA.FILLED = file.path(tmp.analysis, paste(analysis, range, "data", sep = "-"))
        load(file = PATH.DATA.FILLED)
        range.data = range.data[["data"]]
        filled = nrow(range.data$get.mails()) != 0
        if (start == 0 && filled) start = i
        if (start != 0 && filled) end   = i
    }
    ranges.subset = c(
        if (start != 0) rep(FALSE, start - 1),
        rep(TRUE, end - start + 1),
        if (end != -1) rep(FALSE, length(conf$get.entry("ranges")) - end)
    )

    return(ranges.subset)
}



get.filled.ranges = function(casestudy) {
    results = list()
    for (artifact in ARTIFACTS) {
        for (artifact.relation in ARTIFACT.RELATIONS) {
            results[length(results) + 1] = list(get.filled.ranges.for.conf(casestudy, artifact, artifact.relation))
        }
    }

    results.join = join.boolean.values(results, op = "or")

    return(results.join)
}


get.active.ranges = function(casestudy, threshold = c("meansd", "mean", "median")) {
    threshold = match.arg(threshold)

    r = collect.frequencies.from.disk.metrics(casestudy, ranges.subset = NULL)

    thresholds = data.frame()
    results = list()
    for (artifact in ARTIFACTS) {
        for (artifact.relation in ARTIFACT.RELATIONS) {
            for (motif in MOTIFS) {
                data = r[[casestudy]][[artifact.relation]][[artifact]][[ paste(motif, "cr", sep = ".") ]]
                data.threshold = switch(
                    threshold,
                    mean = mean(data),
                    meansd = mean(data) + sd(data),
                    median = median(data)
                )
                ranges.active = data >= data.threshold

                results = c(results, list(ranges.active))

                thresholds = rbind(thresholds, data.frame(
                    artifact = artifact,
                    artifact.relation = artifact.relation,
                    motif = motif,
                    threshold = data.threshold
                ))
            }
        }
    }

    ## set correct order of levels
    thresholds$artifact = factor(thresholds[["artifact"]], levels = ARTIFACTS)

    ## join classifications
    results.join = join.boolean.values(results, op = "or")
    attr(results.join, "thresholds") = thresholds

    return(results.join)
}
