## Claus Hunsen, 2016
## hunsen@fim.uni-passau.de


## / / / / / / / / / / / / / /
## Set environment for script
##

## libraries
library(plyr) # for rbind.fill
library(igraph) # graphs, baby!
library(ggplot2) # plots
library(viridis) # for cool colors
library(limma) # for Venn diagrams
library(svglite) # for svg export via ggsave

## source script files
source(file.path("util.R")) # get our utility functions
source(file.path("data-output.R")) # get our functions related to data output
source(file.path("sensitivity.R")) # get our functions related to sensitivity-data output


## / / / / / / / / / / / / / /
## Frequency evolution
##

plot.frequency.evolution = function(results, bars = FALSE, thresholds = NULL, smooth = TRUE) {

    ## set output path
    plot.path = file.path(TMP.PLOT)
    if(bars)  plot.path = file.path(plot.path, "active")
    dir.create(plot.path, recursive = TRUE, showWarnings = FALSE)

    steps.available = FALSE

    data.plot = data.frame()

    for (ARTIFACT in ARTIFACTS) {
        for (ARTIFACT.RELATION in ARTIFACT.RELATIONS) {

            motif.data = results[[ CASESTUDY ]][[ ARTIFACT.RELATION ]][[ ARTIFACT ]] #$motif.count
            ranges = unique(motif.data[[ "range" ]])
            steps.available = steps.available || "step" %in% colnames(motif.data)

            data.plot.current = lapply(ranges, function(range) {

                ## get range data
                motif.data.current = motif.data[ motif.data[["range"]] == range, ]
                range.current.which = rownames(motif.data.current)[1] # first available row number to use as label
                if (as.integer(range.current.which) < 10)
                    range.current.which = paste0("0", range.current.which)
                range.current = paste0("r", range.current.which)

                to.append = data.frame()

                for (MOTIF in MOTIFS) {

                    appendix.df = data.frame(
                        "range" = range,
                        "range.lab" = range.current,
                        "motif" = MOTIF,
                        "artifact" = ARTIFACT,
                        "artifact.relation" = ARTIFACT.RELATION,
                        "cr" = motif.data.current[[ paste(MOTIF, "collaborating", sep = ".") ]],
                        "fcr" = motif.data.current[[ paste(MOTIF, "collaborating.and.mailing", sep = ".") ]],
                        "fraction" = motif.data.current[[ paste(MOTIF, "p2", sep = ".") ]]
                    )

                    if (steps.available) {
                        appendix.df = cbind(
                            appendix.df,
                            step = motif.data.current[["step"]],
                            step.error = motif.data.current[[ paste(MOTIF, "p2.step.error", sep = ".") ]]
                        )
                    }

                    to.append = rbind.fill(to.append, appendix.df)

                    # if (artifact == artifacts[1]) {
                    #     to.append =
                    #         rbind.fill(to.append,
                    #                    data.frame(
                    #                        "range" = range,
                    #                        "range.lab" = range.current,
                    #                        "motif" = motif,
                    #                        "artifact" = "baseline",
                    #                        "artifact.relation" = artifact.relation,
                    #                        "fraction" = motif.data.current[[ paste(motif, "p1", sep = ".") ]]
                    #                    )
                    #         )
                    # }

                }

                return(to.append)
            })

            data.plot = rbind(data.plot, rbind.fill(data.plot.current))
        }
    }

    ## transform ranges to dates for x-axis ticks
    data.plot$range.date = get.date.from.unix.timestamp(
        sapply(data.plot$range, function(range) {
            bounds = get.range.bounds(range)
            return(bounds[1])
        })
    )

    ## reorder motifs and artifacts properly
    data.plot["artifact"] = factor(
        tools::toTitleCase(as.vector(data.plot[["artifact"]])),
        ARTIFACTS.UPPER
    )
    data.plot["motif"] = factor(
        tools::toTitleCase(as.vector(data.plot[["motif"]])),
        MOTIFS.UPPER
    )

    ## transform step data to factor so it can be used later
    if (steps.available) {
        data.plot["step"] = factor(as.character(data.plot[["step"]]))
    }

    ## compute artifact-independent, i.e., project-level data:
    ## 1) construct formula
    agg.formula.groups = "range + range.lab + range.date + motif + artifact.relation"
    if (steps.available) {
        agg.formula.groups = paste(agg.formula.groups, "+ step")
    }
    agg.formula = paste(agg.formula.groups, "~ .")
    ## 2) re-compute data and construct global data
    data.plot.global.cr = dcast(data.plot, agg.formula, sum, value.var = "cr")
    colnames(data.plot.global.cr)[which(colnames(data.plot.global.cr) == ".")] = "cr"
    data.plot.global.fcr = dcast(data.plot, agg.formula, sum, value.var = "fcr")
    colnames(data.plot.global.fcr)[which(colnames(data.plot.global.fcr) == ".")] = "fcr"
    data.plot.global = merge(data.plot.global.cr, data.plot.global.fcr, sort = FALSE)
    rm(data.plot.global.cr, data.plot.global.fcr)
    # data.plot.global["artifact"] = "overall"
    data.plot.global["fraction"] = data.plot.global[["fcr"]] / data.plot.global[["cr"]]

    if (steps.available) {
        data.plot.global.step.error = dcast(data.plot, agg.formula, mean, value.var = "step.error", na.rm = TRUE)
        colnames(data.plot.global.step.error)[which(colnames(data.plot.global.step.error) == ".")] = "step.error"
        data.plot.global = merge(data.plot.global, data.plot.global.step.error, sort = FALSE)
        rm(data.plot.global.step.error)
    }


    ## dump plot data for inspection ----
    plot.data.save = list(
        empirical = data.plot,
        smooth = data.plot.global
    )
    save(plot.data.save, file = file.path(plot.path, paste0(CASESTUDY, "-data.Rdata")))


    for (MOTIF in MOTIFS.UPPER) {

        dat = data.plot[ data.plot$motif == MOTIF, ]
        dat.global = data.plot.global[ data.plot.global$motif == MOTIF, ]
        threshs = thresholds[ thresholds$motif == MOTIF, ]

        if (tolower(MOTIF) == "triangle") {
            dat = dat[ dat$artifact.relation == "cochange", ]
            dat.global = dat.global[ dat.global$artifact.relation == "cochange", ]
        }


        ## plot fraction of fulfilled coordination requirements (f=fcr/cr) as line plots, facet = c(artifact.relation, motif) ----

        g.f = ggplot(data = dat, aes(x = range.date, y = fraction, group = artifact))

        if (bars) {
            g.f = g.f +
                geom_bar(aes(fill = artifact), stat = "identity",
                         width = 0.75, position = position_dodge(width = 0.75)) +

                scale_fill_viridis(discrete = TRUE, end = 0.85, begin = 0, option = "viridis",
                                   name = "", breaks = levels(dat$artifact))
        } else {

            g.f = g.f +

                # ## global value
                # geom_line(data = dat.global,
                #           aes(group = 1),
                #           size = 1,
                #           color = "red",
                #           linetype = "solid",
                #           inherit.aes = TRUE) +

                ## values for each artifact
                geom_point(aes(shape = artifact, color = artifact), size = 5) +
                geom_line(aes(linetype = artifact, color = artifact), size = 1.5)

            if (smooth) {
                g.f = g.f +

                    ## smoothing for global value
                    geom_smooth(
                        data = dat.global,
                        aes(group = 1, fill = "smooth"),
                        method =  "loess",
                        se = FALSE,
                        color = "#CF4446",
                        linetype = "solid",
                        size = 3.5,
                        span = mean(
                            (7 / length(unique(dat.global$range))) * length(ARTIFACTS),
                            (7 / length(unique(dat.global$range)))
                        )
                    )
            }

            g.f = g.f +

                scale_x_datetime(labels = scales::date_format("%Y"), date_breaks = "2 years") +

                scale_color_viridis(discrete = TRUE, end = 0.85, begin = 0, option = "viridis",
                                    name = "", breaks = levels(dat$artifact)) +
                scale_linetype_manual(name = "", values = c("solid", "longdash", "dashed"),
                                      breaks = levels(dat$artifact)) +
                scale_shape_manual(name = "", values = c(15, 16, 17),
                                   breaks = levels(dat$artifact)) +
                scale_fill_manual(name = "", labels = "Smoothed Value Across Artifacts", values = "#CF4446") +

                guides(color = guide_legend(order = 1),
                       linetype = guide_legend(order = 1),
                       shape = guide_legend(order = 1),
                       fill = guide_legend(order = 2))

                # scale_color_manual(name = "", labels = legend.items, values = legend.colors) +
                # scale_linetype_manual(name = "", values = legend.lty, breaks =  legend.items) +
                # scale_shape_manual(name = "", values = legend.shapes, breaks = legend.items) +
                # scale_fill_manual(name = "", labels = "smoothed value across artifacts", values = "blue")
        }

        g.f = g.f +
            labs(
                # title = paste(CASESTUDY, "-", MOTIF, "motif"),
                # x = "revision ranges",
                x = NULL,
                y = "Fraction of Fulfilled\nCoordination Requirements"
            ) +
            # scale_x_discrete(breaks = NULL, labels = unique(dat$range.lab)) +
            ylim(0, 1)

        if (!steps.available) {
            g.f = g.f +
                facet_grid(motif ~ .)
        } else {
            g.f = g.f +
                facet_grid(step ~ motif)
                # ggforce::facet_grid_paginate(motif ~ step, page = 1)
        }

        g.f = g.f +
            theme_light(base_size = 35) +
            theme(
                # axis.text.x = element_text(angle = 90, vjust = 0.5),
                strip.text = element_text(colour = "black"),
                plot.title = element_text(hjust = 0.5),
                legend.position = "bottom"
            )#  +
            # ggtitle(CASESTUDIES.REAL[[CASESTUDY]])

        plot.width = 16
        plot.height = 8
        if (steps.available) plot.height = plot.height * SENSITIVITY.STEPS.COUNT
        ggsave(plot = g.f,
               filename = file.path(plot.path, paste0(CASESTUDY, "-", tolower(MOTIF), ".pdf")),
               width = plot.width, height = plot.height, limitsize = FALSE)
        # ggsave(plot = g.f,
        #        filename = file.path(plot.path, paste0(CASESTUDY, "-", tolower(MOTIF), ".png")),
        #        width = plot.width, height = 5)
        # ggsave(plot = g.f,
        #        filename = file.path(plot.path, paste0(CASESTUDY, "-", tolower(MOTIF), ".svg")),
        #        width = plot.width, height = 5)
        save(g.f, file = file.path(plot.path, paste0(CASESTUDY, "-", tolower(MOTIF), ".Rdata")))


        ## plot evolution of amount of (fulfilled) coordination requirements (cr and fcr) ----

        dat = reshape2::melt(dat, measure.vars = c("cr", "fcr"))
        # dat.global = reshape2::melt(dat.global, measure.vars = c("cr", "fcr"))

        if (MOTIF == "triangle") {
            dat = dat[ dat$artifact.relation == "cochange", ]
        }

        g.cr.fcr <- ggplot(data = dat, aes(x = range, y = value, group = variable))
        if (bars) {
            g.cr.fcr = g.cr.fcr +
                geom_bar(aes(fill = variable), stat = "identity",
                                           width = 0.75, position = position_dodge(width = 0.75))
        } else {
            g.cr.fcr = g.cr.fcr +
                geom_point(aes(shape = variable, color = variable), size = 5) +
                geom_line(aes(linetype = variable, color = variable), size = 2) # +

                # ## smoothing for global value
                # geom_smooth(
                #     data = dat.global,
                #     aes(linetype = variable, color = variable),
                #     method =  "loess",
                #     linetype = "dashed",
                #     span = mean(
                #         (7 / length(unique(dat.global$range))) * length(ARTIFACTS),
                #         (7 / length(unique(dat.global$range)))
                #     ),
                #     show.legend = FALSE
                # )
        }

        if (!is.null(thresholds)) {
            g.cr.fcr = g.cr.fcr +
                geom_hline(data = threshs, aes(yintercept = threshold), linetype = "dashed", alpha = "0.5")
        }

        if (!steps.available) {
            g.cr.fcr = g.cr.fcr +
                facet_grid(artifact ~ ., scales = "free")
        } else {
            g.cr.fcr = g.cr.fcr +
                facet_grid(step ~ artifact, scales = "free")
            # ggforce::facet_grid_paginate(motif ~ step, page = 1)
        }

        g.cr.fcr = g.cr.fcr +
            labs(
                # title = paste(MOTIF, "motif"), # CASESTUDIES.REAL[CASESTUDY], "-",
                # x = "revision ranges",
                x = NULL,
                y = "Number of (Fulfilled)\nCoordination Requirements" #,
            ) +
            scale_x_discrete(labels = unique(dat$range.lab)) +
            scale_y_log10(labels = scales::comma) +
            scale_linetype_manual(name = "", values = c("longdash", "solid"),
                                  breaks = levels(dat$variable),
                                  labels=c("All Coordination Requirements", "Fulfilled Coordination Requirements")) +

            scale_color_viridis(discrete = TRUE, end = 0.85, begin = 0, option = "viridis",
                                name = "", breaks = levels(dat$variable),
                                labels=c("All Coordination Requirements", "Fulfilled Coordination Requirements")) +
            scale_fill_viridis(discrete = TRUE, end = 0.85, begin = 0, option = "viridis",
                               name = "", breaks = levels(dat$variable),
                               labels=c("All Coordination Requirements", "Fulfilled Coordination Requirements")) +
            # scale_fill_brewer(palette = "Set1", # values = c("skyblue3", "darkgreen", "orangered1"),
            #                   name = "", breaks = levels(dat$variable),
            #                   labels=c("All Coordination Requirements", "Fulfilled Coordination Requirements")) +
            # scale_color_brewer(palette = "Set1", #values = c("orangered1", "darkgreen"),
            #                    name = "", breaks = levels(dat$variable),
            #                    labels=c("All Coordination Requirements", "Fulfilled Coordination Requirements")) +

            scale_shape_manual(values = c(15,16,17),
                               name = "", breaks = levels(dat$variable),
                               labels=c("All Coordination Requirements", "Fulfilled Coordination Requirements")) +
            theme_light(base_size = 17) +
            theme(
                axis.text.x = element_text(angle = 90, vjust = 0.5),
                strip.text = element_text(colour = "black"),
                plot.title = element_text(hjust = 0.5),
                legend.position = "bottom"
            )#  +
            # ggtitle(CASESTUDIES.REAL[[CASESTUDY]])

        plot.width = 16
        plot.height = 5
        if (steps.available) plot.height = plot.height * SENSITIVITY.STEPS.COUNT
        ggsave(plot = g.cr.fcr,
               filename = file.path(plot.path, paste0(CASESTUDY, "-", tolower(MOTIF), "-raw.pdf")),
               width = plot.width, height = plot.height, limitsize = FALSE)
        # ggsave(plot = g.cr.fcr,
        #        filename = file.path(plot.path, paste0(CASESTUDY, "-", tolower(MOTIF), "-raw.png")),
        #        width = plot.width, height = 5)
        # ggsave(plot = g.cr.fcr,
        #        filename = file.path(plot.path, paste0(CASESTUDY, "-", tolower(MOTIF), "-raw.svg")),
        #        width = plot.width, height = 5)
        save(g.cr.fcr, file = file.path(plot.path, paste0(CASESTUDY, "-", tolower(MOTIF), "-raw.Rdata")))


        ## plot relative errors of sensitivity analysis ----

        if (steps.available) {

            g.f = ggplot(data = dat, aes(x = range.date, y = step.error, group = artifact))

            if (bars) {
                g.f = g.f +
                    geom_bar(aes(fill = artifact), stat = "identity",
                             width = 0.75, position = position_dodge(width = 0.75)) +

                    scale_fill_viridis(discrete = TRUE, end = 0.85, begin = 0, option = "viridis",
                                       name = "", breaks = levels(dat$artifact))
            } else {

                g.f = g.f +

                    ## values for each artifact
                    geom_point(aes(shape = artifact, color = artifact), size = 5) +
                    geom_line(aes(linetype = artifact, color = artifact), size = 1.5)

                if (smooth) {
                    g.f = g.f +

                        ## smoothing for global value
                        geom_smooth(
                            data = dat.global,
                            aes(group = 1, fill = "smooth"),
                            method =  "loess",
                            se = FALSE,
                            color = "#CF4446",
                            linetype = "solid",
                            size = 3.5,
                            span = mean(
                                (7 / length(unique(dat.global$range))) * length(ARTIFACTS),
                                (7 / length(unique(dat.global$range)))
                            )
                        )
                }

                g.f = g.f +

                    scale_x_datetime(labels = scales::date_format("%Y"), date_breaks = "2 years") +

                    scale_color_viridis(discrete = TRUE, end = 0.85, begin = 0, option = "viridis",
                                        name = "", breaks = levels(dat$artifact)) +
                    scale_linetype_manual(name = "", values = c("solid", "longdash", "dashed"),
                                          breaks = levels(dat$artifact)) +
                    scale_shape_manual(name = "", values = c(15, 16, 17),
                                       breaks = levels(dat$artifact)) +
                    scale_fill_manual(name = "", labels = "Smoothed Value Across Artifacts", values = "#CF4446") +

                    guides(color = guide_legend(order = 1),
                           linetype = guide_legend(order = 1),
                           shape = guide_legend(order = 1),
                           fill = guide_legend(order = 2))
            }

            g.f = g.f +
                labs(
                    # title = paste(CASESTUDY, "-", MOTIF, "motif"),
                    # x = "revision ranges",
                    x = NULL,
                    y = "Relative Error in\nFraction of Fulfilled Coordination Requirements"
                ) +
                ylim(0, 1) +

                ## grid
                facet_grid(step ~ motif) +

                theme_light(base_size = 35) +
                theme(
                    strip.text = element_text(colour = "black"),
                    plot.title = element_text(hjust = 0.5),
                    legend.position = "bottom"
                )

            plot.width = 16
            plot.height = 8 * SENSITIVITY.STEPS.COUNT
            ggsave(plot = g.f,
                   filename = file.path(plot.path, paste0(CASESTUDY, "-", tolower(MOTIF), "-error.pdf")),
                   width = plot.width, height = plot.height, limitsize = FALSE)
            # ggsave(plot = g.f,
            #        filename = file.path(plot.path, paste0(CASESTUDY, "-", tolower(MOTIF), "-error.png")),
            #        width = plot.width, height = plot.height, limitsize = FALSE)
            ggsave(plot = g.f,
                   filename = file.path(plot.path, paste0(CASESTUDY, "-", tolower(MOTIF), "-error.svg")),
                   width = plot.width, height = plot.height, limitsize = FALSE)
            save(g.f, file = file.path(plot.path, paste0(CASESTUDY, "-", tolower(MOTIF), "-error.Rdata")))
        }
    }

}


## / / / / / / / / / / / / / /
## PLOTTING WRAPPER
##

plot.evolution <- function(ranges.subset = NULL, bars = FALSE, thresholds = NULL) {
    logging::loginfo("Plotting the frequency evolution.")
    results = collect.frequencies.from.disk(CASESTUDY, ranges.subset = ranges.subset)
    plot.frequency.evolution(results, bars = bars, thresholds = thresholds)
}

plot.null.models <- function(conf, ranges.subset = NULL) {
    logging::loginfo("Plotting the null models.")

    ## create path for file
    null.subfolder = sprintf("null-%s", NULL.MODEL)
    if (REMOVE.CORE) {
        null.subfolder = paste0(null.subfolder, "-without-core")
    }
    plot.path = file.path(TMP.PLOT, null.subfolder)
    dir.create(plot.path, recursive = TRUE, showWarnings = FALSE)

    ## configuration file: ranges
    ranges.full = conf$get.entry("ranges")
    ranges = conf$get.entry("ranges")
    if (!is.null(ranges.subset)) ranges = ranges[ ranges.subset ]

    casestudy = CASESTUDY
    results.null = data.frame()
    for(artifact.relation in ARTIFACT.RELATIONS) {
        for (artifact in ARTIFACTS) {
            ## get temporary path for storage of intermediate data
            analysis = construct.analysis.id(CASESTUDY, artifact, artifact.relation)
            tmp.folder = file.path(TMP, CASESTUDY, analysis)
            tmp.folder.null = file.path(tmp.folder, null.subfolder)
            tmp.folder.null.original = file.path(tmp.folder)
            if (REMOVE.CORE) {
                tmp.folder.null.original = file.path(tmp.folder.null, "original")
            }

            logging::loginfo("Collecting null-model data for %s", analysis)

            null.data.path = file.path(plot.path, paste0(analysis, ".Rdata"))
            null.data = save.and.load("null.data", null.data.path, skip = RERUN.ANALYSIS, if.not.found = function() {

                ## get data for each revision range
                null.data = lapply(ranges, function(range) {

                    ## get original data
                    files.original = list.files(
                        tmp.folder.null.original,
                        pattern = paste0(range, "-motifs.data$"),
                        full.names = TRUE
                    )
                    motif.counts = lapply(files.original, function(f) {
                        logging::logdebug("Loading empirical data from file '%s'.", f)
                        load(f)
                        return(motifs.data)
                    })[[1]]

                    ## get null-model files
                    files = list.files(tmp.folder.null, pattern = paste0(range, "-null-.*-motifs.data$"), full.names = TRUE)
                    data.list = lapply(files, function(f) {
                        logging::logdebug("Loading null-model data from file '%s'.", f)
                        load(f)
                        return(motifs.data)
                    })
                    names(data.list) = 1:length(files)
                    logging::logdebug("Combining null-model data.")
                    data = combine.data.to.single.data.frame(data.list)

                    ## add empirical data
                    data["triangle.empirical"] = motif.counts[["triangle"]][["p2"]]
                    data["square.empirical"] = motif.counts[["square"]][["p2"]]

                    ## add range data
                    data["range"] = range

                    range.current.which = which(ranges.full == range)
                    if (as.integer(range.current.which) < 10)
                        range.current.which = paste0("0", range.current.which)
                    range.current = paste0("r", range.current.which)
                    data["range.plot"] = range.current

                    return(data)
                })
                # names(null.data) = ranges
                null.data = rbind.fill(null.data)
                null.data["artifact"] = artifact

                ## subset data
                null.data = null.data[c("range", "range.plot", "artifact", "triangle.p2", "square.p2", "triangle.empirical", "square.empirical")]

                return(null.data)
            })

            ## store in results
            results.null = rbind.fill(results.null, null.data)
        }
    }

    for (MOTIF in MOTIFS) {

        logging::loginfo("Plotting null-model data for motif %s", MOTIF)

        g.null = ggplot(data = results.null, aes_string(x = paste(MOTIF, "p2", sep = "."))) +
            scale_x_continuous(limits = c(0, 1)) +
            geom_histogram(aes(y = ..density..),
                           bins = 30,
                           colour = "black",
                           fill = "white") +
            geom_point(
                aes_string(x = paste(MOTIF, "empirical", sep = ".")),
                y = 0,
                color = "red",
                size = 5
            ) +
            geom_density(alpha = .2, fill = "#AAD4FF") +
            facet_grid(factor(artifact, levels = ARTIFACTS) ~ range.plot, scales = "free") +
            theme_light() +
            scale_color_brewer(palette = "Set1") +
            theme(
                plot.title = element_text(
                    lineheight = .8,
                    face = "bold",
                    size = 16
                ),
                axis.title.x = element_text(size = 14),
                axis.title.y = element_text(size = 14),
                axis.text.x = element_text(angle = 90, vjust = 0.5, hjust = 1),
                legend.title = element_text(lineheight = .8, size = 14),
                legend.text = element_text(size = 12),
                strip.text.x = element_text(colour = "black", size = 12),
                strip.text.y = element_text(colour = "black", size = 12)
            )

        ggsave(plot = g.null, filename = file.path(plot.path, paste0(CASESTUDY, "-", MOTIF, "-null.pdf")),
               width = length(ranges) * 2, height = length(ARTIFACTS) * 4, limitsize = FALSE)
        # ggsave(plot = g.null, filename = file.path(plot.path, paste0(CASESTUDY, "-null.png")), width = length(ranges) * 2, height = length(ARTIFACTS) * 4, limitsize = FALSE)
        # ggsave(plot = g.null, filename = file.path(plot.path, paste0(CASESTUDY, "-null.svg")), width = length(ranges) * 2, height = length(ARTIFACTS) * 4, limitsize = FALSE)

    }

}

plot.sensitivity.models = function(conf, ranges.subset = ranges.subset.filled, thresholds = NULL, smooth = FALSE) {
    logging::loginfo("Plotting the Kossinets evolution.")

    results = collect.sensitivity.frequencies.from.disk(CASESTUDY, ranges.subset = ranges.subset)

    # fix output path for sensitivity plots
    TMP.PLOT.BACKUP = TMP.PLOT
    TMP.PLOT <<- file.path(TMP.PLOT, "sensitivity")

    plot.frequency.evolution(results, bars = FALSE, thresholds = thresholds, smooth = smooth)

    ## restore output path
    TMP.PLOT <<- TMP.PLOT.BACKUP
}


## / / / / / / / / / / / / / /
## Venn diagrams
##

plot.venn.diagrams = function(config, ranges.subset = NULL) {

    construct.venn.diagrams.for.motif = function(dat, motif = "triangle") {
        venn = rbind(
            c(x1 = 0, x2 = 0, Counts = dat[[paste0(motif, ".complete")]]),
            c(1, 0, dat[[paste0(motif, ".collaborating")]]),
            c(0, 1, paste0(dat[[paste0(motif, ".mailing")]], "\n(",
              round(dat[[paste0(motif, ".p1")]], digits = 2), ")")),
            c(1, 1, paste0(dat[[paste0(motif, ".collaborating.and.mailing")]], "\n(", round(dat[[paste0(motif, ".p2")]], digits = 2), ")"))
        )
        attr(venn, "class") = "VennCounts"
        attr(venn, "developers") = dat[[paste0(motif, ".developers")]]

        return(venn)
    }

    logging::loginfo("Plotting the Venn diagrams.")

    ## create path for file
    plot.path = file.path(TMP.PLOT, "venn")
    file = file.path(plot.path, paste0("%s-venn-%s.pdf"))
    dir.create(plot.path, recursive = TRUE, showWarnings = FALSE)

    ## configuration file: ranges
    ranges.full = conf$get.entry("ranges")
    ranges = conf$get.entry("ranges")
    if (!is.null(ranges.subset)) ranges = ranges[ ranges.subset ]

    ## create path for file
    dir = dirname(file)
    dir.create(dir, recursive = TRUE, showWarnings = FALSE)

    results = collect.frequencies.from.disk(CASESTUDY, ranges.subset = ranges.subset)

    for (artifact.relation in ARTIFACT.RELATIONS) {
        for (artifact in ARTIFACTS) {
            analysis = construct.analysis.id(CASESTUDY, artifact, artifact.relation)
            results.artifact = results[[ CASESTUDY ]][[ artifact.relation ]][[ artifact ]]
            for (motif in MOTIFS) {
                file.artifact = sprintf(file, analysis, motif)
                results.artifact.motif = results.artifact[
                    c("range", grep(motif, colnames(results.artifact), fixed = TRUE, value = TRUE))
                ]

                ## open PDF device
                pdf(file = file.artifact, height = 8, width = 6)
                par(mfrow=c(3,2))

                for(range in ranges) {
                    ## get data for range
                    venn.data = results.artifact.motif[results.artifact.motif$range == range, ]
                    venn = construct.venn.diagrams.for.motif(venn.data, motif = motif)
                    ## plot
                    vennDiagram(venn, names = c("mailing", "collaborating", "coll+mail"),
                                cex = 1, counts.col = "red", mar = rep(0, 4)) # limma
                    title(main = range, sub = paste(attr(venn, "developers"), "developers"),
                          line = 1, cex.main = 1.1, cex.sub = 1.1)
                }

                ## close device
                graphics.off()
            }
        }
    }

}
