## Claus Hunsen, 2017
## hunsen@fim.uni-passau.de


## / / / / / / / / / / / / / /
## Set environment for script
##

## libraries
library(BiRewire)

## source script files
source(file.path("util.R")) # get our utility functions
source(file.path("data.R")) # data analysis


## / / / / / / / / / / / / / /
## DATA GATHERING
##

remove.core.authors.from.networks = function(networks, range.data) {

    ## get networks
    authors.to.artifacts = networks[["authors.to.artifacts"]]
    authors.net = networks[["authors.net"]]

    ## to be sure we only analyze the authors for their coreness
    authors.net.authors = igraph::delete.vertices(authors.net, igraph::V(authors.net)[type != TYPE.AUTHOR])

    ## calculate core authors
    core.mail = get.author.class.by.type(network = authors.net.authors, type = "network.degree")[["core"]][["author.name"]]
    core.commit = get.author.class.by.type(data = range.data, type = "commit.count")[["core"]][["author.name"]]
    cores = intersect(core.mail, core.commit)
    # browser(expr = length(cores) > 1)

    ## remove core authors
    authors.net = igraph::delete.vertices(authors.net, cores)
    authors.to.artifacts[which(names(authors.to.artifacts) %in% cores)] = NULL

    ## return re-newed networks list
    networks[["authors.net"]] = authors.net
    networks[["authors.to.artifacts"]] = authors.to.artifacts

    ## add core authors as attribute
    attr(networks, "core") = cores
    logging::loginfo("Remove core authors from author network: [%s].",
                     paste(cores, collapse = ", "))

    return(networks)
}


## get the multi networks (see CodefaceRangeData$get.multi.network())
get.null.model.for.net.builder = function(net.builder, range.data, network.conf) {

    networks = net.builder$get.networks()

    ## remove core authors from network data
    if (REMOVE.CORE) {
        networks = remove.core.authors.from.networks(networks, range.data)
    }

    authors.to.artifacts = networks[["authors.to.artifacts"]]
    authors.net = networks[["authors.net"]]
    artifacts.net = networks[["artifacts.net"]]

    ## unify vertices with author-artifact relation
    authors.from.net = igraph::get.vertex.attribute(authors.net, "name")
    authors.from.artifacts = names(authors.to.artifacts)
    authors.net = authors.net + igraph::vertices(setdiff(authors.from.artifacts, authors.from.net), type = TYPE.AUTHOR)
    ## unify vertices with artifacts from bipartite and artifact relation
    artifacts.all = range.data$get.artifacts()
    artifacts.from.net = igraph::get.vertex.attribute(artifacts.net, "name")
    artifacts.net = artifacts.net + igraph::vertices(setdiff(artifacts.all, artifacts.from.net), type = TYPE.ARTIFACT)

    if (is.directed(authors.net) && !is.directed(artifacts.net)) {
        logging::logwarn("Author network is directed, but artifact network is not. Converting artifact network...")
        artifacts.net = as.directed(artifacts.net, mode = "mutual")
    } else if (!is.directed(authors.net) && is.directed(artifacts.net)) {
        logging::logwarn("Author network is undirected, but artifact network is not. Converting artifact network...")
        artifacts.net = as.undirected(artifacts.net, mode = "collapse", edge.attr.comb = EDGE.ATTR.HANDLING)
    }

    switch(
        NULL.MODEL,
        ## rewire edges by keeping the degree distribution for all author vertices
        rewire = {
            ## rewire author--author network
            if (ecount(authors.net) > 1) {
                logging::loginfo("Rewiring author--author network.")
                authors.net.adj = as.matrix(get.adjacency(authors.net))
                authors.net.adj.rewired = birewire.rewire.undirected(authors.net.adj, verbose = FALSE)
                authors.net.null = graph.adjacency(authors.net.adj.rewired, mode = "undirected", weighted = TRUE)
                ## because we may get multiple edges between two vertices after rewiring, we need to simplify again
                authors.net.null = simplify.network(authors.net.null)
                ## re-set vertex attributes
                authors.net.null = igraph::set.vertex.attribute(authors.net.null, "id", value =
                                                                    igraph::get.vertex.attribute(authors.net, "id"))
                authors.net.null = igraph::set.vertex.attribute(authors.net.null, "type", value = TYPE.AUTHOR)
            } else {
                logging::loginfo("Skipping rewiring of author--author network (no edges or just one).")
                authors.net.null = authors.net
            }

            ## Test degree dist
            if(!all(sort(as.vector(degree(authors.net.null))) == sort(as.vector(degree(authors.net)))))
                logging::logwarn("Degree distribution not conserved during rewiring.")
        },
        random = {
            ## randomize author vertices in author--author network
            if (ecount(authors.net) > 0) {
                logging::loginfo("Randomize author vertices in author--author network.")

                ## randomize vertex names
                authors.names = get.vertex.attribute(authors.net, "name")
                authors.random = sample(authors.names)

                ## set randomized vertex names
                authors.net.null = set.vertex.attribute(authors.net, "name", value = authors.random)
            } else {
                logging::loginfo("Skipping randomizing author vertices in author--author network (no edges).")
                authors.net.null = authors.net
            }
        },
        errg = {
            ## generate Erdös-Renyi network as author--author network
            if (ecount(authors.net) > 0) {
                logging::loginfo("Generate Erdös-Renyi network as author--author network.")

                ## get original network data as data.frames
                original.edges = igraph::as_data_frame(authors.net, what = "edges")
                original.vertices = igraph::as_data_frame(authors.net, what = "vertices")
                original.vertices.names = igraph::get.vertex.attribute(authors.net, "name")

                ## generate random network with original size and get data.frame of it
                g.random = igraph::erdos.renyi.game(igraph::vcount(authors.net), p.or.m = igraph::ecount(authors.net),
                                                    type = "gnm", directed = FALSE, loops = FALSE)
                g.random.edges = igraph::as_data_frame(g.random, what = "edges")

                ## replace original vertices on edges with random ones, plus transform numeric IDs to real vertex names
                edges.random = original.edges
                edges.random["from"] = original.vertices.names[g.random.edges[["from"]]]
                edges.random["to"] = original.vertices.names[g.random.edges[["to"]]]

                ## create null-model network
                authors.net.null = igraph::graph_from_data_frame(edges.random, vertices = original.vertices, directed = FALSE)
            } else {
                logging::loginfo("Skipping generating Erdös-Renyi network (no edges).")
                authors.net.null = authors.net
            }
        }
    )

    u = combine.networks(authors.net.null, artifacts.net, authors.to.artifacts, network.conf)

    return(u)

}


## / / / / / / / / / / / / / /
## WRAPPER
##

calculate.null.model = function(conf, network.conf, number) {

    null.model.data = mclapply(conf$get.entry("ranges"), function(range) {

        logging::loginfo("Processing null model (range=%s, number=%s).", range, number)

        PATH.NULL = file.path(TMP.NULL, paste(ANALYSIS, range, "null", number, "motifs.data", sep = "-"))

        motifs.data = load.and.save("motifs.data", PATH.NULL, skip = RERUN.ANALYSIS, if.not.found = function() {
            ## load the filled range.data object from disk
            PATH.DATA.FILLED = file.path(TMP.ANALYSIS, paste(ANALYSIS, range, "data", sep = "-"))
            load(file = PATH.DATA.FILLED)
            net.builder = range.data[["netbuilder"]]
            range.data = range.data[["data"]]

            ## construct null-model network
            PATH.NULL.NET = file.path(TMP.NULL, paste(ANALYSIS, range, "null", number, "net", sep = "-"))
            net = load.and.save("net", PATH.NULL.NET, skip = RERUN.ANALYSIS, if.not.found = function() {
                net = get.null.model.for.net.builder(net.builder, range.data, network.conf)
                return(net)
            })

            ## get motif data for current network
            motifs.data = get.motif.data.for.network(net, range)
            return(motifs.data)
        })

        logging::loginfo("Finished null model (range=%s, number=%s).", range, number)
        return(motifs.data)

    })

    names(null.model.data) = conf$get.entry("ranges")
    invisible(null.model.data)

}

calculate.null.model.original.for.range.data = function(conf) {

    null.model.data = mclapply(conf$get.entry("ranges"), function(range) {
        PATH.BASIS.DATA = file.path(TMP.ANALYSIS, paste(ANALYSIS, range, "data", sep = "-"))
        PATH.BASIS.NET =  file.path(TMP.ANALYSIS, paste(ANALYSIS, range, "network", sep = "-"))
        PATH.ORIGINAL.DATA = file.path(TMP.NULL.ORIGINAL, paste(ANALYSIS, range, "original", sep = "-"))
        PATH.ORIGINAL.NET = file.path(TMP.NULL.ORIGINAL, paste(ANALYSIS, range, "original.net", sep = "-"))
        PATH.ORIGINAL.CORES = file.path(TMP.NULL.ORIGINAL, paste(ANALYSIS, range, "original.cores", sep = "-"))

        motifs.data = load.and.save("motifs.data", PATH.ORIGINAL.DATA, skip = RERUN.ANALYSIS, if.not.found = function() {
            ## construct original-model network
            net = load.and.save("net", PATH.ORIGINAL.NET, skip = RERUN.ANALYSIS, if.not.found = function() {
                ## load the filled range.data object from disk
                load(file = PATH.BASIS.DATA)
                ## load the network
                load(file = PATH.BASIS.NET)

                ## remove core authors:
                ## (1) create dummy networks list
                networks = list(
                    authors.net = net,
                    authors.to.artifact = list()
                )
                ## (2) remove the core authors
                networks = remove.core.authors.from.networks(networks, range.data[["data"]])
                ## (3) get the original network for null-model comparison
                net = networks[["authors.net"]]

                ## store core authors
                core = attr(networks, "core")
                save(core, file = PATH.ORIGINAL.CORES)

                return(net)
            })

            ## get motif data for current network
            motifs.data = get.motif.data.for.network(net, range)
            return(motifs.data)
        })

        return(motifs.data)
    })

    invisible(null.model.data)
}



## / / / / / / / / / / / / / /
## NULL-MODEL CHECK
##

check.null.model = function(conf) {

    range.results = mclapply(conf$get.entry("ranges"), function(range) {
        logging::loginfo("Checking null model (range=%s).", range)

        ## load original network
        original.path = file.path(TMP.ANALYSIS, paste(ANALYSIS, range, "network", sep = "-"))
        load(file = original.path)
        original = strip.attributes.from.network(net)

        ## load null-model networks
        PATH.NULL.NET = paste(ANALYSIS, range, "null.net", "*", sep = "-")
        files = list.files(path = TMP.NULL, pattern = PATH.NULL.NET)
        files = file.path(TMP.NULL, files)
        files = files

        diffs = lapply(files, function(file) {
            load(file = file)
            net = strip.attributes.from.network(net)
            diff = graph.difference(original, net)
            return(diff)
        })
        diffs = sum(unlist(diffs))
        attr(diffs, "vertices") = igraph::V(original)$name

        logging::loginfo("Finished null model (range=%s).", range)
        return(diffs)
    })
    names(range.results) = conf$get.entry("ranges")
    range.results = unlist(range.results)

    print(range.results)
    print(sprintf("sum = %s", sum(range.results)))

}
