# coding=utf-8
# Claus Hunsen, 2016
# hunsen@fim.uni-passau.de

import itertools
import logging
import os
import socket
import tempfile

from plumbum import local

# configure logging
logging.basicConfig(level=logging.INFO)


#
# HELPER FUNCTIONS
#

def get_container_path(mount, file=""):
    """Transforms path to a path inside the uchroot container."""
    absdir = os.path.abspath(str(mount))
    dirname = os.path.split(absdir)[-1]
    container_path = os.path.join("/mnt", dirname, file)
    return container_path


def get_configuration_name(conf, chunk=None):
    # re-order configuration parts
    conf = conf[1:] + (conf[0],)
    # add chunk information
    if chunk:
        conf = (chunk,) + conf
    # join to string
    return "_".join(conf)


#
# PATHS
#

# basic information
DIR = os.path.dirname(os.path.realpath(__file__))
HOST = socket.gethostname()
PATH_ROOT = "/scratch/codeface"

# Codeface
PATH_ANALYSIS = os.path.join(PATH_ROOT, "network-coordination-requirements")
PATH_ANALYSIS_SCRIPT = "run.sh"

# benchbuild + uchroot container
PATH_BENCHBUILD = os.path.join(PATH_ROOT, "benchbuild")
PATH_CONTAINER = os.path.join(PATH_ROOT, "container")
PATH_CONTAINER_FILE = os.path.join(PATH_CONTAINER, "2018-09-11_R_minimal.tar.bz2")
# PATH_CONTAINER_FILE = os.path.join(PATH_BENCHBUILD, "results", "codeface.oNm0L7r9", "container-in")

#
# SLURM
#

# node

CH_SLURM_ENV = ["-p", "anywhere", "-A", "anywhere", "--exclusive", "--constraint=zeus|chimaira|cayman|pontipine", "--mem=0", "--time=12:0:0"]
#CH_SLURM_ENV = ["-p", "anywhere", "-A", "anywhere", "--exclusive", "--constraint=zeus", "--mem=0", "--time=24:0:0"]
#CH_SLURM_ENV = ["-p", "sphinx", "-A", "sphinx", "--exclusive", "--mem=0", "--time=24:0:0"]
#CH_SLURM_ENV = ["-p", "anywhere", "-A", "anywhere", "--exclusive", "--constraint=zeus|chimaira|cayman|pontipine", "--mem=0", "--qos=verylong", "--time=0"]
#CH_SLURM_ENV = ["-p", "anywhere", "-A", "anywhere", "--exclusive", "--constraint=zeus|chimaira|pontipine", "--mem=0", "--qos=verylong", "--time=0"]

# notification email
USER = "hunsen"
CH_MAIL = "{}@fim.uni-passau.de".format(USER)

# job parameters
CH_SLURM_JOB_NAME = "net-cr"
CH_SLURM_PARAMS = ["--get-user-env"]
CH_SBATCH = os.path.join(PATH_ANALYSIS, "run-container-wrapper.sh")

# set job dependency to singleton (last of user's jobs)
CH_SLURM_DEPENDENCY = ["--dependency=singleton"]  # run after all jobs of user
CH_SLURM_MAIL = ["--mail-type=END", "--mail-user=" + CH_MAIL]

#
#  ANALYSIS CONFIGURATIONS
#

SCRIPT = "data"  # one of: data, null, sensitivity, nullcheck, stats, history-plots
LOG_LEVEL = "INFO"

# list of casestudies
CASESTUDIES = [
    "apache-http",
    "busybox",
    "ffmpeg",
    "git",
    "llvm",
    "openssl",
    "postgres",
    "qemu",
    "uboot",
    "wine",

# # no feature data:
#    "django",
#    "jailhouse",

# # broken somehow
#    "gcc",

# # testing
#    "test",
]

# list of artifacts
ARTIFACTS = [
    "file",
    "function",
    "feature"
]

# list of artifact relations
ARTIFACT_RELATIONS = [
    "cochange"  # ,
    # "callgraph"
]

# the Slurm directive for the null-model job steps
# (make sure that 24 is always run!)
NULL_MODEL_STEPS = "0-24%3"

# time-window length used for data splitting (e.g., "3 months" and "2 weeks")
SPLIT_WINDOW = "3 months"

# use a sliding-window approach for time windows?
SLIDING_WINDOW = False

# null model to use (one of: rewire, random, errg)
NULL_MODEL = "rewire"

# remove core authors from null model?
REMOVE_CORE = False

# selection process (one of: releases, threemonth, testing)
SELECTION_PROCESS = "threemonth"

# rerun analysis completely and ignore dumped intermediate data?
RERUN_ANALYSIS = False

# construct all configurations
all_configurations = {
    'data': [element for element in itertools.product(["data"], CASESTUDIES, ARTIFACTS, ARTIFACT_RELATIONS)],
    'null': [element for element in itertools.product(["null"], CASESTUDIES, ARTIFACTS, ARTIFACT_RELATIONS)],
    'sensitivity': [element for element in itertools.product(["sensitivity"], CASESTUDIES, ARTIFACTS, ARTIFACT_RELATIONS)],
    'nullcheck': [element for element in itertools.product(["nullcheck"], CASESTUDIES, ARTIFACTS, ARTIFACT_RELATIONS)],
    'stats': [element for element in itertools.product(["stats"], [CASESTUDIES[0]], [ARTIFACTS[0]], [ARTIFACT_RELATIONS[0]])],
    'history-plots': [element for element in itertools.product(["history-plots"], CASESTUDIES, [ARTIFACTS[0]], [ARTIFACT_RELATIONS[0]])],
}
configurations = all_configurations[SCRIPT]
# configurations = configurations[:1]
# configurations = configurations[1:]

#
# ANALYSIS PATHS
#

# global paths
CODEFACE_DATA = os.path.join(PATH_ROOT, "codeface-data")
CODEFACE_DATA_CONTAINER = get_container_path(CODEFACE_DATA)

# logs
LOGS_SUBFOLDERS = {
  "threemonth" : SPLIT_WINDOW.replace(' ', '_'),
  "releases" : "releases"
}

LOGS = os.path.join(PATH_ANALYSIS, "logs-v3", LOGS_SUBFOLDERS[SELECTION_PROCESS])  # slurm log
LOGS_CONTAINER = os.path.join(PATH_ANALYSIS, "logs-v3", LOGS_SUBFOLDERS[SELECTION_PROCESS])  # within container

#
# COMMAND
#

# run the command for each configuration
for configuration in configurations:

    # construct case-study-dependent parameters
    LOG_FILE_SLURM_SUFFIX = ("-" + NULL_MODEL) if SCRIPT is "null" else ""
    LOG_FILE_SLURM = os.path.join(
        LOGS,
        configuration[1],
        get_configuration_name(configuration),
        get_configuration_name(configuration) + LOG_FILE_SLURM_SUFFIX + ".slurm"
    )

    # create log folder
    LOG_FILE_PATH = os.path.dirname(LOG_FILE_SLURM)
    os.makedirs(LOG_FILE_PATH, exist_ok=True)

    # construct configuration-dependent parameters
    TMP_FOLDER = tempfile.TemporaryDirectory(prefix="net-cr_container_", dir=os.path.join("/local", os.environ["USER"])).name

    # sbatch wrapper
    sbatch = local["sbatch"][
        CH_SLURM_ENV,
        CH_SLURM_PARAMS,
        "-J", "-".join([CH_SLURM_JOB_NAME, configuration[1], configuration[0]]),
        # CH_SLURM_DEPENDENCY,
        CH_SLURM_MAIL
    ]

    # add array parameter for 'null' script
    if SCRIPT in ["null", "sensitivity"]:
        LOG_FILE_SLURM = LOG_FILE_SLURM.replace(".slurm", "_%a.slurm")
        sbatch = sbatch[
            "--array={}".format(NULL_MODEL_STEPS)
        ]

    # add wrapper script
    sbatch = sbatch[
        "-o", LOG_FILE_SLURM,
        CH_SBATCH
    ]

    # container command
    cmd = sbatch["container",
                 # "--help",
                 "-i", "{}".format(PATH_CONTAINER_FILE),
                 "-m", "{}".format(PATH_ANALYSIS),  # FIXME sync this to cluster:/local/codeface/{tmp}/
                 "-m", "{}".format(CODEFACE_DATA),  # FIXME write on cluster first, then sync result files
                 "-t", TMP_FOLDER,
                 "run",
                 "--",
                 "/bin/bash",
                 get_container_path(PATH_ANALYSIS, PATH_ANALYSIS_SCRIPT),
                 configuration,
                 "-d", get_container_path(CODEFACE_DATA_CONTAINER),
                 "-s", SELECTION_PROCESS,
                 "-w", SPLIT_WINDOW,
                 "--loglevel", LOG_LEVEL
    ]

    if SCRIPT is "null" or SCRIPT is "stats":
        cmd = cmd["--null-model", NULL_MODEL]

    if SLIDING_WINDOW:
        cmd = cmd["--sliding-window"]

    if not REMOVE_CORE:
        cmd = cmd["--no-remove-core"]

    if RERUN_ANALYSIS:
        cmd = cmd["--complete-rerun"]

    # call the command
    with local.cwd(PATH_BENCHBUILD):
        logging.info("Scheduling {} task".format(CH_SLURM_JOB_NAME))
        logging.info("-- configuration: {}".format(get_configuration_name(configuration, SELECTION_PROCESS)))
        logging.info("-- command: {}".format(cmd))
        # cmd()  # (cmd > CH_SLURM_OUTPUT)()
        task_id = ''.join(x for x in cmd() if x.isdigit())
        logging.info("-- task ID: {}".format(task_id))
        logging.info("")
