## Claus Hunsen, 2020
## hunsen@fim.uni-passau.de


## / / / / / / / / / / / / / /
## Set environment for script
##

## source script files
source(file.path("util.R")) # get our utility functions
source(file.path("data.R")) # data analysis
source(file.path("data-output.R")) # data output and reading


SIMULATION.ALGORITHMS = c("NRE", "BSPC", "BSPA")
SENSITIVITY.STEPS = seq(from = 0.1, to = 0.9, by = 0.1)
SENSITIVITY.STEPS.COUNT = length(SENSITIVITY.STEPS) + 1 # steps + original network/data

project.bipartite.network = function(net.bi) {
    net.uni = igraph::bipartite_projection(
        net.bi,
        types = igraph::get.vertex.attribute(net.bi, "type"), which = "true",
        multiplicity = TRUE, remove.type = FALSE
    )
    return(net.uni)
}

simulate.kossinets.model = function(bipartite.network, algo = SIMULATION.ALGORITHMS) {

    algo = match.arg(algo)
    range = attr(bipartite.network, "range")

    ## network properties
    authors = igraph::V(bipartite.network)[type == TRUE]
    artifacts = igraph::V(bipartite.network)[type == FALSE]
    edges = igraph::E(bipartite.network)
    N = length(authors)
    M = length(artifacts)

    ## identify artifacts per author
    artifacts.per.author = lapply(authors, function(author) {
        author.vertex = igraph::V(bipartite.network)[author]
        ## all vertices involved in edges including 'author'
        edge.vertices = igraph::V(bipartite.network)[ .nei(author.vertex) ]
        return(edge.vertices)
    })
    ## identify artifacts with more than one author
    artifacts.with.more.authors = artifacts[table(unlist(artifacts.per.author)) > 1]

    ## run 10-percent steps for all kinds of models
    model.networks = lapply(SENSITIVITY.STEPS, function(step) {

        switch(
            algo,
            BSPC = {
                ## remove random authors
                remove.func = igraph::delete.vertices
                to.remove = sample(authors, as.integer(N * step), replace = FALSE)
            },
            BSPA = {
                ## remove random artifacts
                remove.func = igraph::delete.vertices
                to.remove = sample(artifacts, as.integer(M * step), replace = FALSE)
            },
            NRE = {
                ## remove edges among randomly selected authors within unipartite networks,
                ## i.e., remove edges from selected non-respondent authors to artifacts with more than one author
                remove.func = igraph::delete.edges
                non.respondents = sample(authors, as.integer(N * step), replace = FALSE)
                to.remove = igraph::E(bipartite.network)[ inc(non.respondents) ][ inc(artifacts.with.more.authors) ]
            }
        )

        ## adjust bipartite network
        net = remove.func(bipartite.network, to.remove)
        attr(net, "range") = range
        attr(net, "step") = step
        attr(net, "algo") = algo

        return(net)
    })

    ## clone original network by altering it
    original.network = bipartite.network
    original.network = igraph::set.graph.attribute(original.network, "model", value = "Kossinets")
    original.network = igraph::set.graph.attribute(original.network, "algo", value = algo)
    ## add attributes to original network
    attr(original.network, "range") = range
    attr(original.network, "step") = 0
    attr(original.network, "algo") = algo

    ## construct list of networks to return
    networks = c(
        list(original.network),
        model.networks
    )
    names(networks) = c(0, SENSITIVITY.STEPS)

    return(networks)
}

get.kossinets.model.for.net.builder = function(net.builder, range.data, bipartite.author.net, network.conf) {

    networks = net.builder$get.networks()
    authors.to.artifacts = networks[["authors.to.artifacts"]]
    artifacts.net = networks[["artifacts.net"]]

    ## project given bipartite author network as unipartite network and fix attributes
    authors.net = project.bipartite.network(bipartite.author.net)
    attributes(authors.net) = attributes(bipartite.author.net)
    igraph::V(authors.net)$type = TYPE.AUTHOR
    igraph::E(authors.net)$type = TYPE.EDGES.INTRA

    ## add all missing authors to the network if wanted
    ## as the given bipartite network does not add all authors (due to Kossinets metrics!)
    if (network.conf$get.value("author.all.authors")) {
        authors.from.all = net.builder$get.project.data()$get.authors()[[ "author.name" ]]
        authors.from.net = igraph::get.vertex.attribute(authors.net, "name")
        authors.net = authors.net + igraph::vertices(setdiff(authors.from.all, authors.from.net), type = TYPE.AUTHOR)
    }

    ## unify vertices with author-artifact relation
    authors.from.net = igraph::get.vertex.attribute(authors.net, "name")
    authors.from.artifacts = names(authors.to.artifacts)
    authors.net = authors.net + igraph::vertices(setdiff(authors.from.artifacts, authors.from.net), type = TYPE.AUTHOR)
    ## unify vertices with artifacts from bipartite and artifact relation
    artifacts.all = range.data$get.artifacts()
    artifacts.from.net = igraph::get.vertex.attribute(artifacts.net, "name")
    artifacts.net = artifacts.net + igraph::vertices(setdiff(artifacts.all, artifacts.from.net), type = TYPE.ARTIFACT)

    if (is.directed(authors.net) && !is.directed(artifacts.net)) {
        logging::logwarn("Author network is directed, but artifact network is not. Converting artifact network...")
        artifacts.net = as.directed(artifacts.net, mode = "mutual")
    } else if (!is.directed(authors.net) && is.directed(artifacts.net)) {
        logging::logwarn("Author network is undirected, but artifact network is not. Converting artifact network...")
        artifacts.net = as.undirected(artifacts.net, mode = "collapse", edge.attr.comb = EDGE.ATTR.HANDLING)
    }

    ## combine network
    u = combine.networks(authors.net, artifacts.net, authors.to.artifacts, network.conf)

    ## re-use object attributes
    attributes(u) = attributes(bipartite.author.net)

    return(u)
}

calculate.kossinets.metrics = function(range, bipartite.networks) {

    # bipartite.networks = unlist(bipartite.networks, recursive = FALSE) ## flatten list of networks

    metrics.data = lapply(bipartite.networks, function(net.bi) {

        step = attr(net.bi, "step")
        algo = attr(net.bi, "algo")

        ## construct unipartite network from bipartite one
        net.uni = project.bipartite.network(net.bi)

        ## network properties
        authors = igraph::V(net.bi)[type == TRUE]
        artifacts = igraph::V(net.bi)[type == FALSE]

        ## metrics (bipartite)
        N = length(authors)
        M = length(artifacts)
        mu = mean(igraph::degree(net.bi, authors))
        nu = mean(igraph::degree(net.bi, artifacts))
        r_b = igraph::assortativity.degree(net.bi, directed = FALSE)

        ## helpers (unipartite)
        degrees.uni = igraph::degree(net.uni)
        components.uni = components(net.uni)
        components.uni.largest = igraph::induced.subgraph(
            net.uni,
            which(components.uni$membership == which.max(components.uni$csize))
        )

        ## metrics (unipartite)
        z = mean(degrees.uni)
        V = var(degrees.uni)
        C = igraph::transitivity(net.uni, type = "global", isolates = "NaN")
        r_u = igraph::assortativity.degree(net.uni, directed = FALSE)
        N_C = components.uni$no
        S_L = max(components.uni$csize)
        l_L = igraph::mean_distance(net.uni, directed = FALSE, unconnected = TRUE)

        ## calculate metrics

        return(data.frame(
            ## context
            range = range,
            algo = algo,
            step = step,
            ## bipartite
            N = N,
            M = M,
            mu = mu,
            nu = nu,
            r_b = r_b,
            ## unipartite
            z = z,
            V = V,
            C = C,
            r_u = r_u,
            N_C = N_C,
            S_L = S_L,
            l_L = l_L
        ))
    })
    metrics.data = rbind.fill(metrics.data)

    return(metrics.data)

}

## / / / / / / / / / / / / / /
## SAVE DATA
##

save.kossinets.data = function(metrics, algo, type = c("kossinets", "motifs"),
                               base.path = TMP.SENSITIVITY, analysis.suffix = NULL) {

    type = match.arg(type)

    analysis.string = ANALYSIS
    analysis.string = paste(analysis.string, type, algo, sep = "-")
    if (!is.null(analysis.suffix))
        analysis.string = paste(analysis.string, analysis.suffix, sep = "-")
    attr(results, "analysis") = analysis.string

    ## dump results to disk
    ## 1) path
    save.path = file.path(base.path, analysis.string)
    ## 2) raw data
    # save(metrics, file = save.path)
    ## 3) CSV
    write.csv(metrics, file = paste0(save.path, ".csv"), row.names = FALSE, fileEncoding = "utf-8")

}

collect.sensitivity.frequencies.from.disk = function(casestudy, ranges.subset = NULL, cols = NULL, cols.rename = NULL) {

    results = list()
    for(artifact.relation in ARTIFACT.RELATIONS) {
        for (artifact in ARTIFACTS) {
            ## get temporary path for storage of intermediate data
            analysis = construct.analysis.id(casestudy, artifact, artifact.relation)
            tmp.folder = file.path(TMP, casestudy, analysis, "sensitivity")
            data.file = file.path(tmp.folder, paste0(analysis, "-motifs", ".csv"))

            ## skip file if not existing
            if (!file.exists(data.file)) {
                next
            }

            ## read data
            data = read.frequencies.from.disk(data.file)

            ## subset columns
            if (!is.null(cols)) {
                data = data[cols]
                if (!is.null(cols.rename))
                    colnames(data) = cols.rename
            }

            ## remove ranges
            if (!is.null(ranges.subset)) {
                data = data[ rep(ranges.subset, times = SENSITIVITY.STEPS.COUNT), ]
            }

            ## store in results
            results[[ casestudy ]][[ artifact.relation ]][[ artifact ]] = data
        }
    }

    return(results)
}


## / / / / / / / / / / / / / /
## WRAPPER
##

construct.kossinets.networks = function(conf, number, algo) {

    algo = match.arg(algo, SIMULATION.ALGORITHMS)

    range.networks = mclapply(conf$get.entry("ranges"), function(range) {

        TMP.NETWORKS = file.path(TMP.SENSITIVITY, paste(ANALYSIS, range, number, algo, "networks", sep = "-"))
        networks = load.and.save("networks", TMP.NETWORKS, skip = RERUN.ANALYSIS, if.not.found = function() {

            ## load original ProjectData (range.data)
            PATH.DATA = file.path(TMP.ANALYSIS, paste(ANALYSIS, range, "data", sep = "-"))
            load(file = PATH.DATA)
            ## load original NetworkBuilder (net.builder)
            PATH.NETBUILDER = file.path(TMP.ANALYSIS, paste(ANALYSIS, range, "netbuilder", sep = "-"))
            load(file = PATH.NETBUILDER)
            ## change NetworkConf temproarily for constructing mail-based bipartite network
            net.builder$set.network.conf.entry("artifact.relation", AUTHOR.RELATION) # use author relation in bipartite network
            net.builder$set.network.conf.entry("author.all.authors", FALSE)  # only use authors from this relation to obtain proper Kossinets metrics
            ## construct original email-based bipartite network (change NetworkConf!)
            original.bipartite = net.builder$get.bipartite.network()
            original.bipartite.vertex.types = igraph::get.vertex.attribute(original.bipartite, "type") == TYPE.AUTHOR
            ## change vertex attribute 'type' to boolean values
            original.bipartite = igraph::remove.vertex.attribute(original.bipartite, "type")
            if (igraph::vcount(original.bipartite) > 0) { # only if more than zero vertices – to not lose the attribute!
                original.bipartite = igraph::set.vertex.attribute(original.bipartite, "type",
                                                                  value = original.bipartite.vertex.types)
            } else {
                value = 0
                class(value) = "logical"
                original.bipartite = igraph::set.vertex.attribute(original.bipartite, "type", value = value)
                original.bipartite = igraph::set.vertex.attribute(original.bipartite, "type", value = NA)
            }

            ## construct simulation networks
            networks = simulate.kossinets.model(original.bipartite, algo)

            return(networks)
        })

        return(networks)
    })

    ## return networks invisibly
    invisible(range.networks)

}

calculate.kossinets.model = function(conf, number, algo) {

    TMP.METRICS = file.path(TMP.SENSITIVITY, paste(ANALYSIS, "kossinets", algo, number, sep = "-"))
    metrics = load.and.save("metrics", TMP.METRICS, skip = RERUN.ANALYSIS, if.not.found = function() {

        null.model.metrics = mclapply(conf$get.entry("ranges"), function(range) {

            logging::loginfo("Processing Kossinets model (range=%s, number=%s).", range, number)

            ## load the networks from disk ('networks')
            TMP.NETWORKS = file.path(TMP.SENSITIVITY, paste(ANALYSIS, range, number, algo, "networks", sep = "-"))
            load(file = TMP.NETWORKS)

            ## calculate metrics
            metrics.data = calculate.kossinets.metrics(range, networks)
            return(metrics.data)

        })

        metrics = rbind.fill(null.model.metrics)
        return(metrics)
    })

    return(metrics)
}

calculate.motif.metrics.for.kossinets.models = function(conf, network.conf, number, algo) {

    TMP.METRICS = file.path(TMP.SENSITIVITY, paste(ANALYSIS, "motifs", algo, number, sep = "-"))
    motifs.data = load.and.save("motifs.data", TMP.METRICS, skip = RERUN.ANALYSIS, if.not.found = function() {

        null.model.metrics = mclapply(conf$get.entry("ranges"), function(range) {

            logging::loginfo("Processing Kossinets model (range=%s, number=%s).", range, number)

            ## load the filled range.data object from disk
            PATH.DATA.FILLED = file.path(TMP.ANALYSIS, paste(ANALYSIS, range, "data", sep = "-"))
            load(file = PATH.DATA.FILLED)
            net.builder = range.data[["netbuilder"]]
            range.data = range.data[["data"]]

            ## load the (bipartite) author networks from disk ('networks')
            TMP.NETWORKS = file.path(TMP.SENSITIVITY, paste(ANALYSIS, range, number, algo, "networks", sep = "-"))
            load(file = TMP.NETWORKS)

            ## construct null-model network
            PATH.NETS.FULL = file.path(TMP.SENSITIVITY, paste(ANALYSIS, range, number, algo, "networks", "cr", sep = "-"))
            nets.cr.full = load.and.save("nets.cr.full", PATH.NETS.FULL, skip = RERUN.ANALYSIS, if.not.found = function() {
                ## construct complete coordination-requirement network for each step in Kossinets model
                nets.cr.full = lapply(networks, function(bipartite.author.net) {
                    net = get.kossinets.model.for.net.builder(net.builder, range.data, bipartite.author.net, network.conf)
                    return(net)
                })

                return(nets.cr.full)
            })

            motifs.data.steps = lapply(nets.cr.full, function(step.net) {
                ## get motif data for current network
                motifs.data = get.motif.data.for.network(step.net, range)

                return(motifs.data)
            })

            ## combine data and re-arrange a bit
            motifs.data.range = combine.data.to.single.data.frame(motifs.data.steps)
            motifs.data.range["range"] = NULL # remove autogenerated column
            motifs.data.range = cbind(
                range = range,
                step = names(motifs.data.steps),
                algo = algo,
                motifs.data.range
            )

            return(motifs.data.range)
        })

        motifs.data = rbind.fill(null.model.metrics)
        return(motifs.data)
    })

    return(motifs.data)

}

