## Claus Hunsen, 2017
## hunsen@fim.uni-passau.de


## / / / / / / / / / / / / / /
## Set environment for script
##

## libraries
library(xtable)
library(effsize)
library(fractaldim)
library(plyr)

## source script files
source(file.path("util.R")) # get our utility functions
source(file.path("data-motifs.R")) # get motif counts on preprocessed data
source(file.path("data-output.R")) # output data


## / / / / / / / / / / / / / /
## ANALYSIS
##

get.casestudy.print = function(cs) {
    return(sprintf("\\tool{%s}", cs))
}

get.artifact.print = function(art) {
    return(sprintf("\\artifact{%s}", art))
}

get.table = function(casestudy, ranges.subset = NULL, null.model = FALSE) {

    logging::loginfo("Collecting the casestudy table for %s", casestudy)

    results = list()
    for(artifact.relation in ARTIFACT.RELATIONS) {
        for (artifact in ARTIFACTS) {

            ## get temporary path for storage of intermediate data
            analysis = construct.analysis.id(casestudy, artifact, artifact.relation)
            tmp.folder = file.path(TMP, casestudy, analysis)

            ## construct data object
            path.conf = file.path(tmp.folder, paste(analysis, "conf", sep = "-"))
            load(path.conf) # load conf object
            project.data = ProjectData$new(conf)

            ## get path to current data file
            if (null.model) {
                null.subfolder = sprintf("null-%s", NULL.MODEL)
                if (REMOVE.CORE) {
                    null.subfolder = paste0(null.subfolder, "-without-core")
                }
                tmp.folder = file.path(tmp.folder, null.subfolder)
            }
            data.file = file.path(tmp.folder, paste0(analysis, ".csv"))

            ## read data
            data = read.frequencies.from.disk(data.file)

            if (!is.null(ranges.subset)) {
                data = data[ ranges.subset, ]
            }

            tab = data.frame(casestudy = get.casestudy.print(CASESTUDIES.REAL[[casestudy]]))

            ## get project meta-data
            commits = project.data$get.commits.filtered()
            dates = as.POSIXct(commits$date)
            artifacts = project.data$get.artifacts()
            authors = project.data$get.authors()
            emails = project.data$get.mails()
            tab = cbind(tab,
                        time = paste(
                            format(min(dates), "%Y"),
                            format(max(dates), "%Y"),
                            sep = "--"),
                        commits = length(unique(commits$commit.id)),
                        emails = length(unique(emails$message.id)),
                        # ranges = nrow(data),
                        developers = nrow(authors),
                        artifact = get.artifact.print(artifact),
                        artifacts = length(artifacts))

            ## get motif data
            for (motif in MOTIFS) {
                cols = paste0(motif, c(".collaborating", ".collaborating.and.mailing"))

                dat = data[cols]
                dat = as.data.frame(t(colSums(dat)))
                dat[[paste0(motif, ".f")]] = dat[[cols[2]]] / dat[[cols[1]]]

                colnames(dat) = paste0(motif, c(".cr", ".fcr", ".f"))
                tab = cbind(tab, dat)
            }

            ## store in results
            results = append(results, list(tab))
        }
    }

    results = rbind.fill(results)
    return(results)

}

print.latex.table = function(df) {

    plot.path = file.path(TMP.STATS)
    dir.create(plot.path, recursive = TRUE, showWarnings = FALSE)
    tex.file = file.path(plot.path, "projects.tex")

    strCaption <- paste0("List of subject projects, combined with their empirical data on fulfilled coordination requirements.")

    duplicated.cols = 1:5
    duplicated.cols = names(df)[duplicated.cols]

    rle.lengths <- rle(df[["casestudy"]])$lengths
    first <- !duplicated(df[["casestudy"]])

    for (i in duplicated.cols) {
        df[[i]][!first] <- ""
        # define appearance of \multirow
        if (i %in% c("commits", "emails", "developers"))
            df[[i]][first] = paste0("\\tablenum{", df[[i]][first], "}")
        df[[i]][first] <-
            paste0("{\\multirow{", rle.lengths, "}{*}{", df[[i]][first], "}}")
    }

    length.artifact = length(unique(df$artifact))
    length.casestudy = length(unique(df$casestudy)) - 1 # remove "case study" '""'!

    ## construct separators:
    separators.pos = list()
    separators.command = c()
    ## 1) top rule + headings + mid rule
    separators.pos = c(separators.pos, list(-1))
    separators.command = c(separators.command,
                           paste(
                               "\\toprule\n",  # NEW row
                               paste(c("\\bfseries Project",
                                       "\\bfseries Time",
                                       "\\bfseries \\# Commits",
                                       "\\bfseries \\# E-Mails",
                                       # "\\multirow{2}{5em}{\\centering\\bfseries \\# Revision Ranges}",
                                       "\\bfseries \\# Developers",
                                       "\\bfseries Artifact",
                                       "\\bfseries \\# Artifacts",
                                       "\\multicolumn{3}{c}{\\bfseries Triangle motif $\\motiftriangle$}",
                                       "\\multicolumn{3}{c}{\\bfseries Square motif $\\motifsquare$}\\\\\n"),
                                     collapse = " & "),
                               "\\cmidrule(l){8-10} \\cmidrule(l){11-13}\n",
                               "&&&&&&& $\\statcrident$ & $\\statcrfull$ & $\\statcrfrac$ & $\\statcrident$ & $\\statcrfull$ & $\\statcrfrac$ \\\\\n",
                               "\\midrule \n"
                           ))
    ## 2) separators between case studies
    if (length.casestudy > 1) {
        pos = seq(from = length.artifact, to = length.artifact * (length.casestudy - 1), by = length.artifact)
        separators.pos = c(separators.pos, pos)
        separators.command = c(separators.command, rep(paste("\\midrule\n"), times = length.casestudy - 1))
    }
    ## 3) bottom rule
    separators.pos = c(separators.pos, nrow(df))
    separators.command = c(separators.command, paste("\\bottomrule\n"))

    ## set up xtable output
    tab.tex = print(xtable(df, # digits = c(0, 0, 0), # first zero "represents" row numbers which we skip later
                           align = "llllllllllllll", # align and put a vertical line (first "l" again represents column of row numbers)
                           caption = strCaption, label = "tab:projects"),
                    file = tex.file,
                    floating.environment = "sidewaystable",
                    table.placement = "p!",
                    size = "scriptsize", #Change size; useful for bigger tables "normalsize" "footnotesize"
                    include.rownames = FALSE, #Don't print rownames
                    include.colnames = FALSE, #We create them ourselves
                    caption.placement = "top", #"top", NULL
                    hline.after = NULL, #We don't need hline; we use booktabs
                    floating = TRUE, # whether \begin{Table} should be created (TRUE) or not (FALSE)
                    sanitize.text.function = force, # Important to treat content of first column as latex function
                    booktabs = TRUE,
                    add.to.row = list(
                        pos = separators.pos,
                        command = separators.command
                    )
    )
    tab.tex = gsub("begin\\{tabular\\}\\{lllllllllllll\\}",
                   "sisetup{table-format=6.0, table-text-alignment=right}\n\\\\begin{tabular}{ll ccc | rS[table-format=6] S[table-format=6]S[table-format=5 ]S[table-format=1.2] S[table-format=7]S[table-format=7]S[table-format=1.2]}",
                   tab.tex)

    ## remove trailing spaces
    tab.tex = gsub("[[:space:]]*\\n", "\n", tab.tex)

    cat(tab.tex, file = tex.file)

}

get.table.averages.dev = function(casestudy) {

    logging::loginfo("Collecting the casestudy table for %s", casestudy)
    artifact = ARTIFACTS[1]

    results = list()
    for(artifact.relation in ARTIFACT.RELATIONS) {

        ## get temporary path for storage of intermediate data
        analysis = construct.analysis.id(casestudy, artifact, artifact.relation)
        tmp.folder = file.path(TMP, casestudy, analysis)
        data.file = file.path(tmp.folder, paste0(analysis, ".csv"))

        ## construct data object
        path.conf = file.path(tmp.folder, paste(analysis, "conf", sep = "-"))
        load(path.conf) # load conf object
        project.data = ProjectData$new(conf)

        tab = data.frame(
            casestudy = get.casestudy.print(CASESTUDIES.REAL[[casestudy]])
            # artifact.relation = artifact.relation
        )

        ## get project meta-data
        commits = project.data$get.commits()

        ## format function for mean and sd
        format.number = function(num) {
            return( round(num, digits = 2) )
        }

        ## commits per dev
        commits.per.dev.data = unique(commits[c("author.name", "hash")])
        commits.per.dev.data.grouped = split(commits.per.dev.data, commits.per.dev.data["author.name"])
        commits.per.dev = unlist(lapply(commits.per.dev.data.grouped, nrow))
        commits.per.dev = data.frame(
            dev.sum = length(unique(commits.per.dev.data[["hash"]])),
            dev.mean = paste0(format.number(mean(commits.per.dev)), " +- ", format.number(sd(commits.per.dev))),
            dev.median = median(commits.per.dev),
            dev.quantile80 = quantile(commits.per.dev, probs = 0.8, names = FALSE),
            # dev.min = min(commits.per.dev),
            dev.max = max(commits.per.dev)
        )
        ## store in temporary results
        tab = cbind(tab, commits.per.dev)
        rm(commits.per.dev)

        ## store in results
        results = append(results, list(tab))
    }

    results = rbind.fill(results)
    return(results)

}


get.table.averages.artifacts = function(casestudy) {

    logging::loginfo("Collecting the casestudy table for %s", casestudy)
    artifact = ARTIFACTS[1]

    results = list()
    for(artifact.relation in ARTIFACT.RELATIONS) {
        for (artifact in ARTIFACTS) {

            ## get tempo    rary path for storage of intermediate data
            analysis = construct.analysis.id(casestudy, artifact, artifact.relation)
            tmp.folder = file.path(TMP, casestudy, analysis)
            data.file = file.path(tmp.folder, paste0(analysis, ".csv"))

            ## construct data object
            path.conf = file.path(tmp.folder, paste(analysis, "conf", sep = "-"))
            load(path.conf) # load conf object
            project.data = ProjectData$new(conf)

            tab = data.frame(
                casestudy = get.casestudy.print(CASESTUDIES.REAL[[casestudy]])
                # artifact.relation = artifact.relation
            )

            ## get project meta-data
            commits = project.data$get.commits()

            ## format function for mean and sd
            format.number = function(num) {
                return( round(num, digits = 2) )
            }

            ## commits per artifact
            commits.per.artifact.data = unique(commits[c("artifact", "artifact.type", "hash")])
            commits.per.artifact.data = commits.per.artifact.data[commits.per.artifact.data[["artifact"]] != "", ]
            commits.per.artifact.data.grouped = split(commits.per.artifact.data, commits.per.artifact.data["artifact"])
            commits.per.artifact = unlist(lapply(commits.per.artifact.data.grouped, nrow))
            commits.per.artifact = data.frame(
                artifact = get.artifact.print(artifact),
                art.mean = paste0(format.number(mean(commits.per.artifact)), " +- ", format.number(sd(commits.per.artifact))),
                art.median = median(commits.per.artifact),
                art.quantile80 = quantile(commits.per.artifact, probs = 0.8, names = FALSE),
                # art.min = min(commits.per.artifact),
                art.max = max(commits.per.artifact)
            )
            ## store in temporary results
            tab = cbind(tab, commits.per.artifact)
            rm(commits.per.artifact)

            ## store in results
            results = append(results, list(tab))
        }
    }

    results = rbind.fill(results)
    return(results)

}

print.latex.table.averages = function(df, file.suffix = "") {

    plot.path = file.path(TMP.STATS)
    dir.create(plot.path, recursive = TRUE, showWarnings = FALSE)
    tex.file = file.path(plot.path, sprintf("projects-averages-%s.tex", file.suffix))

    strCaption <- paste(sprintf("Commits per %s in subject projects", file.suffix),
                        "(mean, standard deviation, 80\\,\\% quantile, maximum).")

    length.artifact = length(unique(df$artifact))
    length.casestudy = length(unique(df$casestudy))

    duplicated.cols = names(df)[1]
    rle.lengths = rle(df[["casestudy"]])$lengths
    first = !duplicated(df[["casestudy"]])
    if (unique(rle.lengths) != 1) {
        for (i in duplicated.cols) {
            df[[i]][!first] = ""
            # define appearance of \multirow
            df[[i]][first] = paste0("\\multirow{", rle.lengths, "}{*}{", df[[i]][first], "}")
        }
    }

    second.col.head = ifelse(file.suffix == "developer", "\\bfseries \\# Commits", "\\bfseries Artifact")
    col.types = ifelse(file.suffix == "developer", "l S S[table-format=4.2(1)] S S S", "l l S[table-format=4.2(1)] S S S")

    ## construct separators:
    separators.pos = list()
    separators.command = c()
    ## 1) top rule + headings + mid rule
    separators.pos = c(separators.pos, list(-1))
    separators.command = c(separators.command,
                           paste(
                               "\\toprule\n",  # NEW row
                               paste(c("\\bfseries Project",
                                       second.col.head,
                                       sprintf("\\multicolumn{4}{c}{\\bfseries \\# Commits per %s} \\\\\n", file.suffix)
                                       # "\\multicolumn{4}{c}{\\bfseries \\# Commits per artifact}",
                                      ),
                                     collapse = " & "),
                               "\\cmidrule(l){3-6} \n",
                               paste(c(
                                   "",
                                   "",
                                   "\\bfseries Avg. $\\pm$ Std.\\ dev.",
                                   "\\bfseries Median",
                                   "\\bfseries .8 quantile",
                                   "\\bfseries Max. \\\\\n"
                               ), collapse = " & "),
                               "\\midrule \n"
                           ))
    ## 2) separators between case studies
    if (file.suffix  == "artifact") {
        pos = seq(from = length.artifact, to = length.artifact * (length.casestudy - 1), by = length.artifact)
        separators.pos = c(separators.pos, pos)
        separators.command = c(separators.command, rep(paste("\\midrule\n"), times = length.casestudy - 1))
    }
    ## 3) bottom rule
    separators.pos = c(separators.pos, nrow(df))
    separators.command = c(separators.command, paste("\\bottomrule\n"))

    ## set up xtable output
    tab.tex = print(xtable(df, # digits = c(0, 0, 0), # first zero "represents" row numbers which we skip later
                           align = paste0(rep("l", ncol(df) + 1), collapse = ""), # align and put a vertical line (first "l" again represents column of row numbers)
                           caption = strCaption, label = sprintf("tab:projects:averages:%s", file.suffix)),
                    file = tex.file,
                    floating.environment = "table",
                    table.placement = "htp",
                    size = "scriptsize", #Change size; useful for bigger tables "normalsize" "footnotesize"
                    include.rownames = FALSE, #Don't print rownames
                    include.colnames = FALSE, #We create them ourselves
                    caption.placement = "top", #"top", NULL
                    hline.after = NULL, #We don't need hline; we use booktabs
                    floating = TRUE, # whether \begin{Table} should be created (TRUE) or not (FALSE)
                    sanitize.text.function = force, # Important to treat content of first column as latex function
                    booktabs = TRUE,
                    add.to.row = list(
                        pos = separators.pos,
                        command = separators.command
                    )
    )
    tab.tex = gsub(
        paste0("begin\\{tabular\\}\\{", paste(rep("l", ncol(df)), collapse = ""), "\\}"),
        sprintf("def\\\\arraystretch{1.2}\n\\\\sisetup{table-format=6.0, table-text-alignment=right}\n\\\\begin{tabular}{%s}", col.types),
        tab.tex)

    ## remove trailing spaces
    tab.tex = gsub("[[:space:]]*\\n", "\n", tab.tex)

    cat(tab.tex, file = tex.file)

}


collect.results = function(casestudy, ranges.subset = NULL, null.model = FALSE) {
    results = list()
    for(artifact.relation in ARTIFACT.RELATIONS) {
        for (artifact in ARTIFACTS) {

            ## get temporary path for storage of intermediate data
            analysis = construct.analysis.id(casestudy, artifact, artifact.relation)
            tmp.folder = file.path(TMP, casestudy, analysis)
            if (null.model) {
                null.subfolder = sprintf("null-%s", NULL.MODEL)
                if (REMOVE.CORE) {
                    null.subfolder = paste0(null.subfolder, "-without-core")
                }
                tmp.folder = file.path(tmp.folder, null.subfolder)
            }
            data.file = file.path(tmp.folder, paste0(analysis, ".csv"))

            ## read data
            data = read.frequencies.from.disk(data.file)
            data = cbind(casestudy = CASESTUDIES.REAL[[casestudy]],
                         artifact = factor(artifact, levels = ARTIFACTS),
                         data)

            ## remove ranges if wanted
            if (!is.null(ranges.subset)) {
                data = data[ ranges.subset, ]
            }

            ## get motif data
            for (motif in MOTIFS) {
                cols = paste0(motif, c(".collaborating", ".collaborating.and.mailing"))

                dat = data[cols]
                dat = as.data.frame(t(colSums(dat)))
                dat[[paste0(motif, ".f")]] = dat[[cols[2]]] / dat[[cols[1]]]

                colnames(dat) = paste0(motif, c(".cr", ".fcr", ".f"))
                data = cbind(data, dat)
            }

            results = c(results, list(data))
        }
    }
    results = rbind.fill(results)
    return(results)
}


significance.level = 0.05

### pairwise mann-whitney test returning W-values
# (adjusted copy of pairwise.wilcox.test)

pairwise.mann.whitney.test.w = function (x, g, p.adjust.method = p.adjust.methods, paired = FALSE, ...) {

    p.adjust.method <- match.arg(p.adjust.method)
    DNAME <- paste(deparse(substitute(x)), "and", deparse(substitute(g)))
    g <- factor(g)
    METHOD <- if (paired)
        "Wilcoxon signed rank"
    else "Wilcoxon rank sum"
    compare.levels.w <- function(i, j) {
        xi <- x[as.integer(g) == i]
        xj <- x[as.integer(g) == j]
        wilcox.test(xi, xj, paired = paired, ...)$statistic
    }
    WVAL <- pairwise.table(compare.levels.w, levels(g), p.adjust.method)
    ans <- list(method = METHOD, data.name = DNAME, p.value = WVAL)
    class(ans) = "pairwise.htest"
    ans
}

### pairwise mann-whitney test returning adjusted p-values
# (adjusted copy of pairwise.wilcox.test)
pairwise.mann.whitney.test.p = function (x, g, p.adjust.method = p.adjust.methods, paired = FALSE, ...) {

    p.adjust.method <- match.arg(p.adjust.method)
    DNAME <- paste(deparse(substitute(x)), "and", deparse(substitute(g)))
    g <- factor(g)
    METHOD <- if (paired)
        "Wilcoxon signed rank test"
    else "Wilcoxon rank sum test"
    compare.levels.p <- function(i, j) {
        xi <- x[as.integer(g) == i]
        xj <- x[as.integer(g) == j]
        wilcox.test(xi, xj, paired = paired, ...)$p.value
    }
    PVAL <- pairwise.table(compare.levels.p, levels(g), p.adjust.method)
    ans <- list(method = METHOD, data.name = DNAME, p.value = PVAL,
                p.adjust.method = p.adjust.method)
    class(ans) = "pairwise.htest"
    ans
}

get.effect.sizes = function(data, data.group.col, test.result.w, test.result.p) {

    # get (interesting) p-values
    test.value = test.result.w$p.value
    test.p.values = test.result.p$p.value

    # construct pairs with significant results
    significants = which(test.p.values < significance.level, arr.ind = TRUE)
    significants.count = nrow(significants)

    # adjust significance level
    adjusted.sl = significance.level / significants.count

    # get significant pairs and their effect size
    pairs = NULL
    test.effectsize = NULL
    if (significants.count > 0) {
        for(i in seq_len(significants.count)) {

            # get group numbers
            treatments = unique(unlist(rev(dimnames(test.result.p$p.value))))
            treat.id = significants[i, 1] + 1
            treat.name = treatments[treat.id]
            control.id = significants[i, 2]
            control.name = treatments[control.id]

            # paste printable pairs
            pair = paste(treat.name, control.name, sep = "-")
            pairs = c(pairs, pair)

            # get data of significant groups
            treat = data[data[[data.group.col]] == treat.name, "value"]
            control = data[data[[data.group.col]] == control.name, "value"]

            # calculate cliff's delta
            d = cliff.delta(d = treat, f = control, conf.level = (1 - adjusted.sl))$estimate

            # paste cliff's deltas
            test.effectsize = c(test.effectsize, d)
        }
        test.pairs = paste(pairs, collapse = ", ")
    } else {
        test.pairs = NULL
    }


    return(list(
        pairs = test.pairs,
        effects = test.effectsize
    ))
}


##
## HYPOTHESIS 1.1 – Data collection and calculation without null-values
##

get.statistics.hypo1.collect = function(subset = NULL, null.model = FALSE) {

    logging::loginfo("Collecting statistics for hypothesis H1 (data collection)")

    ## set output path
    stats.path = file.path(TMP.STATS, "hypo1-collect", if (null.model) NULL.STRING else "empirical")
    dir.create(stats.path, recursive = TRUE, showWarnings = FALSE)
    stats.path.file = file.path(stats.path, paste0("hypo1-collect", if(!is.null(subset)) paste0("-", subset), ".txt"))

    ## 1) construct melted data.frame:
    data.df = lapply(CASESTUDIES, function(casestudy) {

        TMP.RANGES.FILLED = file.path(TMP, casestudy, "ranges", paste(casestudy, "ranges.filled", sep = "-"))
        load(TMP.RANGES.FILLED)

        ## get active ranges if wanted
        if (!is.null(subset) && subset == "active") {
            TMP.RANGES.ACTIVE = file.path(TMP, casestudy, "ranges", paste(casestudy, "ranges.active", sep = "-"))
            load(TMP.RANGES.ACTIVE)
            thresholds = attr(ranges.subset.active, "thresholds")
            ranges.subset.filled = join.boolean.values(list(
                ranges.subset.filled, ranges.subset.active
            ), op = "and")
        }

        results = collect.results(casestudy, ranges.subset.filled, null.model = null.model)

        ## reduce results to casestudy level
        cols = c("casestudy", "artifact")
        for (motif in MOTIFS) {
            cols = c(cols, paste0(motif, c(".cr", ".fcr", ".f")))
        }
        results = unique(results[cols])

        ## construct melted data.frame
        results.melt = lapply(MOTIFS, function(motif) {
            ## melt data.frame
            cols = paste0(motif, c(".f")) # ".cr", ".fcr",
            melted = melt(results, id.vars = c("casestudy", "artifact"), measure.vars = cols)
            ## add motif column
            melted = cbind(melted, motif = factor(motif, levels = MOTIFS))
            ## remove motif prefix from variable
            melted["variable"] = gsub(paste0(motif, "."), "", melted[["variable"]])

            return(melted)
        })
        results.melt = rbind.fill(results.melt)

        results.melt["artifact"] = factor(
            tools::toTitleCase(as.vector(results.melt[["artifact"]])),
            ARTIFACTS.UPPER
        )
        results.melt["motif"] = factor(
            tools::toTitleCase(as.vector(results.melt[["motif"]])),
            MOTIFS.UPPER
        )

        return(results.melt)
    })
    data.df = rbind.fill(data.df)

    ## 1.1) dump data to disk

    data.df.disk = dcast(data.df, "casestudy + motif ~ artifact")
    write.table(
        data.df.disk, file = file.path(stats.path, paste0("hypo1-collect", if(!is.null(subset)) paste0("-", subset), ".csv")),
        sep = ";", row.names = FALSE
    )

    ## 2) plots:

    ## 2.1) violin plot
    p.vio = ggplot(data.df, aes(x = artifact, y = value, color = artifact)) +
        ylim(0, 1) +
        stat_summary(fun.y = mean, fun.ymin = mean, fun.ymax = mean, geom = "crossbar", size = 1.5 / 2) +
        geom_violin(draw_quantiles = FALSE, size = 1.5, fill = "transparent") +
        geom_jitter(height = 0, width = 0.1, size = 3.5) +
        facet_grid(. ~ motif) +
        ## colors
        viridis::scale_color_viridis(discrete = TRUE, end = 0.85, begin = 0, option = "viridis", name = "") +
        ## labels
        xlab(NULL) + # xlab("Motifs and Artifacts") +
        ylab("Fraction of Fullfilled\nCoordination Requirements") +
        ## theme
        theme_light(
            base_size = 17
        ) +
        theme(
            legend.position = "none",
            strip.text = element_text(color = "black")
        )

    ggsave(file.path(stats.path, paste0("hypo1-collect-violin-plots", if(!is.null(subset)) paste0("-", subset), ".pdf")),
           p.vio, width = 10, height = 5)
    # ggsave(file.path(stats.path, paste0("hypo1-collect-violin-plots", if(!is.null(subset)) paste0("-", subset), ".png")),
    #        p.vio, width = 10, height = 5)
    # ggsave(file.path(stats.path, paste0("hypo1-collect-violin-plots", if(!is.null(subset)) paste0("-", subset), ".svg")),
    #        p.vio, width = 10, height = 5)

    ## 2.2) bar plot
    p.bar = ggplot(data.df, aes(x = casestudy, y = value, fill = artifact)) +
        ylim(0, 1) +
        geom_bar(stat = "identity", color = "black", position = position_dodge(.85), width = .65) +
        facet_grid(motif ~ .) +
        ## colors
        viridis::scale_fill_viridis(discrete = TRUE, end = 0.85, begin = 0, option = "viridis", name = "Artifact",
                                    breaks = ARTIFACTS.UPPER) +
        ## labels
        geom_text(aes(label = gsub("^0.", ".", format(signif(value, digits = 2), digits = 2))), vjust = -0.3, color = "black",
                  position = position_dodge(0.9), size = 3) +
        xlab("Subject Projects") +
        ylab("Fraction of Fullfilled\nCoordination Requirements") +
        ## theme
        theme_light(
            base_size = 17
        ) +
        theme(
            # legend.position = "bottom",
            # legend.direction = "horizontal",
            axis.text.x = element_text(angle = 45, hjust = 1),
            strip.text = element_text(color = "black")
        )

    ggsave(file.path(stats.path, paste0("hypo1-collect-bar-plots", if(!is.null(subset)) paste0("-", subset), ".pdf")),
           p.bar, width = 10, height = 5)
    # ggsave(file.path(stats.path, paste0("hypo1-collect-bar-plots", if(!is.null(subset)) paste0("-", subset), ".png")),
    #        p.bar, width = 10, height = 5)
    # ggsave(file.path(stats.path, paste0("hypo1-collect-bar-plots", if(!is.null(subset)) paste0("-", subset), ".svg")),
    #        p.bar, width = 10, height = 5)

    ## 3) statistics

    ## dump data to disk
    data.df.disk = dcast(data.df, "casestudy + motif ~ artifact")
    write.table(
        data.df.disk, file = file.path(stats.path, paste0("hypo1-collect", if(!is.null(subset)) paste0("-", subset), ".csv")),
        sep = ";", row.names = FALSE
    )

    ## run a t-test for each artifact on triangle motif
    cat("", file = stats.path.file, append = FALSE, fill = FALSE)
    results.stats = lapply(MOTIFS.UPPER, function(motif) {

        for (artifact in ARTIFACTS.UPPER) {
            cat("####", artifact, motif, file = stats.path.file, append = TRUE, fill = TRUE)

            # motif = "triangle"
            df = data.df.disk[data.df.disk[["motif"]] == motif, artifact]

            ## run statistical test
            df.test = t.test(df)
            ## put result in output file
            capture.output( print(df.test, print.gap = 3), file = stats.path.file, append = TRUE)

            ## run statistical test
            df.test.shapiro = shapiro.test(df)
            ## put result in output file
            capture.output( print(df.test.shapiro, print.gap = 3), file = stats.path.file, append = TRUE)
        }

    })

}


##
## HYPOTHESIS 1.1 – Calculation *with* null-values
##

get.statistics.hypo1 = function(subset = NULL, null.model = FALSE) {

    logging::loginfo("Collecting the statistical test for hypothesis H1 (calculation with null-values)")

    stats.path = file.path(TMP.STATS, "hypo1", if (null.model) NULL.STRING else "empirical")
    dir.create(stats.path, recursive = TRUE, showWarnings = FALSE)
    stats.path.file = file.path(stats.path, paste0("hypo1", if(!is.null(subset)) paste0("-", subset), ".txt"))

    ## 1) Read data from disk
    stats.path.input.empirical = file.path(
        TMP.STATS, "hypo1-collect", "empirical", paste0("hypo1-collect", if(!is.null(subset)) paste0("-", subset), ".csv")
    )
    data.empirical = read.table(stats.path.input.empirical, sep = ";", header = TRUE)
    data.empirical.melt = reshape2::melt(data.empirical, id.vars = c("casestudy", "motif"), variable.name = "artifact")
    data.empirical.melt = cbind(data.empirical.melt, origin = "Empirical\nData")

    stats.path.input.null = file.path(
        TMP.STATS, "hypo1-collect", NULL.STRING, paste0("hypo1-collect", if(!is.null(subset)) paste0("-", subset), ".csv")
    )
    data.null = read.table(stats.path.input.null, sep = ";", header = TRUE)
    data.null.melt = reshape2::melt(data.null, id.vars = c("casestudy", "motif"), variable.name = "artifact")
    data.null.melt = cbind(data.null.melt, origin = "Null\nModel")

    ## combine data for plotting
    data.complete = rbind.fill(data.empirical.melt, data.null.melt)


    ## 2) plots

    ## 2.1) violin plot (complete)
    p.vio = ggplot(data.complete, aes(x = factor(origin), y = value)) +
        ylim(0, 1) +
        geom_violin(draw_quantiles = c(0.5), size = 1) +
        geom_jitter(aes(color = factor(casestudy), shape = factor(casestudy)),
                    size = 3.5, height = 0, width = 0.1) +
        facet_grid(motif ~ artifact) +
        ## colors
        viridis::scale_color_viridis(discrete = TRUE, end = 0.85, begin = 0, option = "viridis", name = "Subject Projects") +
        scale_shape_manual(values = seq(from = 1, to = length(unique(data.complete$casestudy)), by = 1),
                           name = "Subject Projects") +
        ## labels
        xlab(NULL) + # xlab("Artifacts") +
        ylab("Fraction of Fullfilled\nCoordination Requirements") +
        ## theme
        theme_light(
            base_size = 17
        ) +
        theme(
            legend.position = "bottom",
            legend.direction = "horizontal",
            axis.text.x = NULL, # element_text(angle = 45, hjust = 1),
            strip.text = element_text(color = "black")
        )

    ggsave(file.path(stats.path, paste0("hypo1-violin-plots", if(!is.null(subset)) paste0("-", subset), ".pdf")),
           p.vio, width = 9, height = 6)
    # ggsave(file.path(stats.path, paste0("hypo1-violin-plots", if(!is.null(subset)) paste0("-", subset), ".png")),
    #        p.vio, width = 8, height = 5)
    # ggsave(file.path(stats.path, paste0("hypo1-violin-plots", if(!is.null(subset)) paste0("-", subset), ".svg")),
    #        p.vio, width = 8, height = 5)


    ## 3) statistics

    ## dump data to disk
    data.df.disk = dcast(data.complete, "casestudy + motif + artifact ~ origin")
    colnames(data.df.disk) = c("casestudy", "motif", "artifact", "empirical data", "null model")
    write.table(
        data.df.disk, file = file.path(stats.path, paste0("hypo1", if(!is.null(subset)) paste0("-", subset), ".csv")),
        sep = ";", row.names = FALSE
    )

    ## run a t-test for each artifact on triangle motif
    cat("", file = stats.path.file, append = FALSE, fill = FALSE)
    results.stats = lapply(MOTIFS.UPPER, function(motif) {

        ## get data for current motif
        data.empirical.subset = data.empirical.melt[data.empirical.melt[["motif"]] == motif, ]
        data.null.subset = data.null.melt[data.null.melt[["motif"]] == motif, ]

        for (artifact in ARTIFACTS.UPPER) {
            df.emp.art = data.empirical.subset[data.empirical.subset[["artifact"]] == artifact, ]
            df.null.art = data.null.subset[data.null.subset[["artifact"]] == artifact, ]

            test.data = data.frame(
                value = c(df.emp.art$value, df.null.art$value),
                grouping = factor(
                    rep(c("emp", "null"), times = c(nrow(df.emp.art), nrow(df.null.art))),
                    levels = c("null", "emp")
                )
            )

            ## run statistical test
            w = pairwise.mann.whitney.test.w(test.data$value, test.data$grouping,
                                             p.adjust.method = "none", paired = TRUE,
                                             exact = FALSE, alternative = "greater")
            p = pairwise.mann.whitney.test.p(test.data$value, test.data$grouping,
                                             p.adjust.method = "fdr", paired = TRUE,
                                             exact = FALSE, alternative = "greater")
            effect.sizes = get.effect.sizes(test.data, "grouping", w, p)

            ## put result in output file
            cat("####", artifact, motif, file = stats.path.file, append = TRUE, fill = TRUE)
            capture.output( print(w$p.value, print.gap = 3), file = stats.path.file, append = TRUE)
            cat("", file = stats.path.file, append = TRUE, fill = TRUE)
            capture.output( print(p$p.value, print.gap = 3), file = stats.path.file, append = TRUE)
            cat("", file = stats.path.file, append = TRUE, fill = TRUE)
            capture.output( print(effect.sizes, print.gap = 3), file = stats.path.file, append = TRUE)

        }
    })
}


##
## HYPOTHESIS 1.2
##

get.statistics.hypo2 = function(subset = NULL, null.model = FALSE) {

    logging::loginfo("Collecting the statistical test for hypothesis H2")

    ## set output path
    stats.path = file.path(TMP.STATS, "hypo2", if (null.model) NULL.STRING else "empirical")
    dir.create(stats.path, recursive = TRUE, showWarnings = FALSE)
    stats.path.file = file.path(stats.path, paste0("hypo2", if(!is.null(subset)) paste0("-", subset), ".txt"))


    ## 1) construct data.frame
    data.df = lapply(CASESTUDIES, function(casestudy) {

        TMP.RANGES.FILLED = file.path(TMP, casestudy, "ranges", paste(casestudy, "ranges.filled", sep = "-"))
        load(TMP.RANGES.FILLED)

        ## get active ranges if wanted
        if (!is.null(subset) && subset == "active") {
            TMP.RANGES.ACTIVE = file.path(TMP, casestudy, "ranges", paste(casestudy, "ranges.active", sep = "-"))
            load(TMP.RANGES.ACTIVE)
            thresholds = attr(ranges.subset.active, "thresholds")
            ranges.subset.filled = join.boolean.values(list(
                ranges.subset.filled, ranges.subset.active
            ), op = "and")
        }

        results = collect.results(casestudy, ranges.subset.filled, null.model = null.model)
        artifacts = factor(results[["artifact"]], levels = ARTIFACTS)

        ## reduce results to casestudy level
        cols = c("casestudy", "artifact")
        for (motif in MOTIFS) {
            cols = c(cols, paste0(motif, c(".cr", ".fcr", ".f")))
        }
        results = unique(results[cols])

        ## construct melted data.frame
        results.melt = lapply(MOTIFS, function(motif) {
            ## melt data.frame
            cols = paste0(motif, c(".f")) # ".cr", ".fcr",
            melted = melt(results, id.vars = c("casestudy", "artifact"), measure.vars = cols)
            ## add motif column
            melted = cbind(melted, motif = factor(motif, levels = MOTIFS))
            ## remove motif prefix from variable
            melted["variable"] = gsub(paste0(motif, "."), "", melted[["variable"]])

            return(melted)
        })
        results.melt = rbind.fill(results.melt)

        results.melt["artifact"] = factor(
            tools::toTitleCase(as.vector(results.melt[["artifact"]])),
            ARTIFACTS.UPPER
        )
        results.melt["motif"] = factor(
            tools::toTitleCase(as.vector(results.melt[["motif"]])),
            MOTIFS.UPPER
        )

        return(results.melt)
    })
    data.df = rbind.fill(data.df)


    ## 2) plots

    ## 2.1) violin plot (complete)
    p.vio = ggplot(data.df, aes(x = motif, y = value)) +
        ylim(0, 1) +
        stat_summary(fun.y = mean, fun.ymin = mean, fun.ymax = mean, geom = "crossbar", size = 1.5 / 2) +
        geom_violin(draw_quantiles = FALSE, size = 1.5, fill = "transparent") +
        geom_jitter(aes(color = artifact, shape = artifact), size = 3.5, height = 0, width = 0.1) +
        ## colors
        viridis::scale_color_viridis(discrete = TRUE, end = 0.85, begin = 0, option = "viridis", name = "Artifacts") +
        scale_shape(solid = TRUE, name = "Artifacts") +
        ## labels
        xlab(NULL) + # xlab("Motifs") +
        ylab("Fraction of Fullfilled\nCoordination Requirements") +
        ## theme
        theme_light(
            base_size = 17
        ) +
        theme(
            legend.position = "right"
        )

    ggsave(file.path(stats.path, paste0("hypo2-violin-plots", if(!is.null(subset)) paste0("-", subset), ".pdf")),
           p.vio, width = 8, height = 5)
    # ggsave(file.path(stats.path, paste0("hypo2-violin-plots", if(!is.null(subset)) paste0("-", subset), ".png")),
    #        p.vio, width = 8, height = 5)
    # ggsave(file.path(stats.path, paste0("hypo2-violin-plots", if(!is.null(subset)) paste0("-", subset), ".svg")),
    #        p.vio, width = 8, height = 5)

    ## 2.2) violin plot (per artifact)
    p.vio2 = ggplot(data.df, aes(x = motif, y = value, color = artifact)) +
        ylim(0, 1) +
        stat_summary(fun.y = mean, fun.ymin = mean, fun.ymax = mean, geom = "crossbar", size = 1.5 / 2) +
        geom_violin(draw_quantiles = FALSE, size = 1.5, fill = "transparent") +
        geom_jitter(height = 0, width = 0.1, size = 3.5) +
        facet_grid(. ~ artifact) +
        ## colors
        viridis::scale_color_viridis(discrete = TRUE, end = 0.85, begin = 0, option = "viridis", name = "") +
        ## labels
        xlab(NULL) + # xlab("Artifacts and Motifs") +
        ylab("Fraction of Fullfilled\nCoordination Requirements") +
        ## theme
        theme_light(
            base_size = 17
        ) +
        theme(
            legend.position = "none",
            strip.text = element_text(color = "black")
        )

    ggsave(file.path(stats.path, paste0("hypo2-violin-plots-by-artifact", if(!is.null(subset)) paste0("-", subset), ".pdf")),
           p.vio2, width = 10, height = 5)
    # ggsave(file.path(stats.path, paste0("hypo2-violin-plots-by-artifact", if(!is.null(subset)) paste0("-", subset), ".png")),
    #        p.vio2, width = 10, height = 5)
    # ggsave(file.path(stats.path, paste0("hypo2-violin-plots-by-artifact", if(!is.null(subset)) paste0("-", subset), ".svg")),
    #        p.vio2, width = 10, height = 5)

    ## 3) statistics

    ## dump data to disk
    data.df.disk = dcast(data.df, "casestudy + artifact ~ motif")
    write.table(
        data.df.disk, file = file.path(stats.path, paste0("hypo2", if(!is.null(subset)) paste0("-", subset), ".csv")),
        sep = ";", row.names = FALSE
    )

    ## run statistical test
    w = pairwise.mann.whitney.test.w(data.df[["value"]], factor(data.df$motif, levels = MOTIFS.UPPER),
                                     p.adjust.method = "none", paired = TRUE, exact = FALSE, alternative = "less")
    p = pairwise.mann.whitney.test.p(data.df[["value"]], factor(data.df$motif, levels = MOTIFS.UPPER),
                             p.adjust.method = "fdr", paired = TRUE, exact = FALSE, alternative = "less")
    effect.sizes = get.effect.sizes(data.df, "motif", w, p)

    ## put result in output file
    capture.output( print(w$p.value, print.gap = 3), file = stats.path.file, append = FALSE)
    cat("", file = stats.path.file, append = TRUE, fill = TRUE)
    capture.output( print(p$p.value, print.gap = 3), file = stats.path.file, append = TRUE)
    cat("", file = stats.path.file, append = TRUE, fill = TRUE)
    capture.output( print(effect.sizes, print.gap = 3), file = stats.path.file, append = TRUE)


    ## 4) Intersection of cr_ident

    ## for all case studies
    cr.casestudies = lapply(CASESTUDIES, function(casestudy) {
        logging::logdebug("H2: %s:", casestudy)

        cr.motifs.path = file.path(stats.path, sprintf("%s-motifs.data.Rdata", casestudy))
        cr.motifs = save.and.load("cr.motifs", cr.motifs.path, skip = RERUN.ANALYSIS, if.not.found = function() {
            ## for all motifs
            cr.motifs = lapply(MOTIFS, function(motif) {

                cr.artifacts.path = file.path(stats.path, sprintf("%s-%s-motifs.data-raw.Rdata", casestudy, motif))
                cr.artifacts = save.and.load("cr.artifacts", cr.artifacts.path, skip = RERUN.ANALYSIS, if.not.found = function() {
                    ## for all artifacts
                    cr.artifacts = parallel::mclapply(ARTIFACTS, function(ARTIFACT) {
                        logging::logdebug("H2: %s: %s: %s", casestudy, ARTIFACT, motif)

                        ## compute files and folders to load from disk:
                        ## 1) case-study folder
                        folder = file.path(
                            TMP,
                            casestudy,
                            construct.analysis.id(casestudy, ARTIFACT, ARTIFACT.RELATION)
                        )
                        ## 2) all raw identified motifs
                        list.motifs.raw = list.files(folder, sprintf("*-motifs.data-raw-%s", motif), full.names = TRUE)
                        ## 3) the ranges to analyze
                        ranges.active.file = file.path(TMP, casestudy, "ranges", sprintf("%s-ranges.filled", casestudy))
                        load(ranges.active.file)
                        list.motifs.raw = list.motifs.raw[ranges.subset.filled]

                        ## for all ranges of case study
                        cr.ranges = lapply(list.motifs.raw, function(motifs.file) {
                            logging::logdebug("H2: %s: %s: %s: %s", casestudy, ARTIFACT, motif, motifs.file)

                            ## find network in which the motifs have been identified (to have names on the vertices!)
                            net.file = gsub(sprintf("-motifs.data-raw-%s", motif), "-network", motifs.file)

                            ## load results
                            load(motifs.file)
                            ## load corresponding networks
                            load(net.file)

                            ## return all identified motifs
                            return(list(
                                identified = unique(lapply(raw$collaborating, function(x) names(x[1:2]))),
                                fulfilled = unique(lapply(raw$collaborating.and.mailing, function(x) names(x[1:2])))
                            ))
                        })

                        return(cr.ranges)
                    })
                    names(cr.artifacts) = ARTIFACTS.UPPER

                    return(cr.artifacts)
                }) ## end: save.and.load("cr.artifacts", ...)

                ## compute the overlap/set-diff among all sets of identified coordination requirements
                logging::logdebug("H2: Computing overlaps for casestudy '%s'...", casestudy)
                overlaps = do.call(
                    function(File, Function, Feature) {
                        as = parallel::mcmapply(
                            File, Function, Feature,
                            FUN = function(range.file, range.function, range.feature) {
                                ## list parameters and their number
                                arts = ls()
                                dim.arts = length(arts)

                                ## collect data (remove artifacts from coordination requirements and only use string names!)
                                ## (names of list are the parameter names!):
                                ## 1) identified CRs
                                data.ident = list(
                                    range.file = range.file$identified,
                                    range.function = range.function$identified,
                                    range.feature = range.feature$identified
                                )
                                ## 2) add combinations of two artifacts to data
                                data.unions = list(
                                    range.file = unique(c(data.ident$range.function, data.ident$range.feature)),
                                    range.function = unique(c(data.ident$range.file, data.ident$range.feature)),
                                    range.feature = unique(c(data.ident$range.file, data.ident$range.function))
                                )
                                ## 3) fulfilled CRs
                                data.fulfilled = list(
                                    range.file = range.file$fulfilled,
                                    range.function = range.function$fulfilled,
                                    range.feature = range.feature$fulfilled
                                )

                                ## as setdiff is asymmetric, we need to compute all combinations and
                                ## compare the sets of cr_ident:
                                ## 1) create all combinations of artifacts and their combinations
                                combs = expand.grid(arts, c(arts, "combined")) # generally all combinations
                                combs = combs[order(combs$Var1, combs$Var2), ] # order row-wise
                                ## 2) compute all set diffs for all combinations
                                setdiffs = apply(combs, 1, function(row) {
                                    x = row[[1]]
                                    y = row[[2]]
                                    ## proper row and col names
                                    x.name = stringr::str_to_title(gsub("range.", "", x))
                                    y.name = stringr::str_to_title(gsub("range.", "", y))

                                    ## assert diagonal
                                    if (x == y) return(data.frame(x = x.name, y = y.name, ident = NA, full = NA))

                                    ## get single/individual data sets
                                    x.data = data.ident[[x]]
                                    if (y == "combined") {
                                        y.data = data.unions[[x]]
                                    } else {
                                        y.data = data.ident[[y]]
                                    }
                                    ## compute setdiff
                                    diff = setdiff(x.data, y.data)
                                    diff.size = length(diff)

                                    ## check how many CRs from 2) have been fulfilled
                                    diff.full = diff[diff %in% data.fulfilled[[x]]]
                                    diff.full.size = length(diff.full)

                                    ## construct result
                                    return(data.frame(x = x.name, y = y.name, ident = diff.size, full = diff.full.size))
                                })
                                ## 3) combine result
                                setdiffs = rbind.fill(setdiffs)

                                ## transform data to a matrix with artifacts as dimension:
                                ## 1) basic form
                                result = stats::reshape(setdiffs, idvar = "y", timevar = "x", direction = "wide")
                                ## 2) order rows
                                row.order = match(c(ARTIFACTS.UPPER, "Combined"), c(sort(ARTIFACTS.UPPER), "Combined"))
                                result = result[row.order, ]
                                ## 3) order columns
                                col.order = c(1, 4:5, 6:7, 2:3)
                                result = result[col.order]
                                ## 4) transform first column into row names (for Reduce to work later)
                                rownames(result) = result[[1]]
                                result[1] = NULL

                                ## return
                                return(result)

                            }, SIMPLIFY = FALSE)
                        return(as)
                    },
                    args = cr.artifacts
                )

                ## add everything up
                logging::logdebug("H2: Reducing results for motif '%s'", motif)
                combined.overlaps = Reduce(`+`, overlaps)
                combined.overlaps = as.data.frame(combined.overlaps)

                return(combined.overlaps)
            }) ## end: cr.motifs = lapply(...)

            names(cr.motifs) = MOTIFS
            return(cr.motifs)
        }) ## end: save.and.load("cr.motifs", ...)

        return(cr.motifs)
    })
    names(cr.casestudies) = CASESTUDIES

    # put result in output file
    setdiffs.file = file.path(stats.path, paste0("hypo2-setdiffs", if(!is.null(subset)) paste0("-", subset), ".txt"))
    old.width = options("width")$width; options(width = 400)
    capture.output(print(cr.casestudies), file = setdiffs.file)
    options(width = old.width)
}

##
## HYPOTHESIS 02
##

get.statistics.hypo3 = function(subset = NULL, null.model = FALSE) {

    logging::loginfo("Collecting the statistical test for hypothesis H3")

    ## set output path
    stats.path = file.path(TMP.STATS, "hypo3", if (null.model) NULL.STRING else "empirical")
    dir.create(stats.path, recursive = TRUE, showWarnings = FALSE)
    stats.path.file = file.path(stats.path, paste0("hypo3", if(!is.null(subset)) paste0("-", subset), ".txt"))

    data.df = lapply(CASESTUDIES, function(casestudy) {

        TMP.RANGES.FILLED = file.path(TMP, casestudy, "ranges", paste(casestudy, "ranges.filled", sep = "-"))
        load(TMP.RANGES.FILLED)

        ## get active ranges if wanted
        if (!is.null(subset) && subset == "active") {
            TMP.RANGES.ACTIVE = file.path(TMP, casestudy, "ranges", paste(casestudy, "ranges.active", sep = "-"))
            load(TMP.RANGES.ACTIVE)
            thresholds = attr(ranges.subset.active, "thresholds")
            ranges.subset.filled = join.boolean.values(list(
                ranges.subset.filled, ranges.subset.active
            ), op = "and")
        }

        results = collect.results(casestudy, ranges.subset.filled, null.model = null.model)
        artifacts = factor(results[["artifact"]], levels = ARTIFACTS)

        ## reduce results to casestudy level
        cols = c("casestudy", "artifact")
        for (motif in MOTIFS) {
            cols = c(cols, paste0(motif, c(".cr", ".fcr", ".f")))
        }
        results = unique(results[cols])

        ## construct melted data.frame
        results.melt = lapply(MOTIFS, function(motif) {
            ## melt data.frame
            cols = paste0(motif, c(".f")) # ".cr", ".fcr",
            melted = melt(results, id.vars = c("casestudy", "artifact"), measure.vars = cols)
            ## add motif column
            melted = cbind(melted, motif = factor(motif, levels = MOTIFS))
            ## remove motif prefix from variable
            melted["variable"] = gsub(paste0(motif, "."), "", melted[["variable"]])

            return(melted)
        })
        results.melt = rbind.fill(results.melt)

        results.melt["artifact"] = factor(
            tools::toTitleCase(as.vector(results.melt[["artifact"]])),
            ARTIFACTS.UPPER
        )
        results.melt["motif"] = factor(
            tools::toTitleCase(as.vector(results.melt[["motif"]])),
            MOTIFS.UPPER
        )

        return(results.melt)
    })
    data.df = rbind.fill(data.df)


    ## 2) plots

    ## 2.1) violin plot (complete)
    p.vio = ggplot(data.df, aes(x = artifact, y = value)) +
        ylim(0, 1) +
        stat_summary(fun.y = mean, fun.ymin = mean, fun.ymax = mean, geom = "crossbar", size = 1.5 / 2) +
        geom_violin(draw_quantiles = FALSE, size = 1.5, fill = "transparent") +
        geom_jitter(aes(color = motif, shape = motif),
                    size = 3.5, height = 0, width = 0.1) +
        ## colors
        viridis::scale_color_viridis(discrete = TRUE, end = 0.85, begin = 0, option = "viridis", name = "Motifs") +
        scale_shape(solid = TRUE, name = "Motifs") +
        ## labels
        xlab(NULL) + # xlab("Artifacts") +
        ylab("Fraction of Fullfilled\nCoordination Requirements") +
        ## theme
        theme_light(
            base_size = 17
        ) +
        theme(
            legend.position = "right"
        )

    ggsave(file.path(stats.path, paste0("hypo3-violin-plots", if(!is.null(subset)) paste0("-", subset), ".pdf")),
           p.vio, width = 8, height = 5)
    # ggsave(file.path(stats.path, paste0("hypo3-violin-plots", if(!is.null(subset)) paste0("-", subset), ".png")),
    #        p.vio, width = 8, height = 5)
    # ggsave(file.path(stats.path, paste0("hypo3-violin-plots", if(!is.null(subset)) paste0("-", subset), ".svg")),
    #        p.vio, width = 8, height = 5)


    ## 3) statistics

    ## dump data to disk
    data.df.disk = dcast(data.df, "casestudy + motif ~ artifact")
    write.table(
        data.df.disk, file = file.path(stats.path, paste0("hypo3", if(!is.null(subset)) paste0("-", subset), ".csv")),
        sep = ";", row.names = FALSE
    )

    ## run statistical test
    w = pairwise.mann.whitney.test.w(data.df[["value"]], factor(data.df$artifact, levels = ARTIFACTS.UPPER),
                                     p.adjust.method = "none", paired = TRUE, exact = FALSE, alternative = "greater")
    p = pairwise.mann.whitney.test.p(data.df[["value"]], factor(data.df$artifact, levels = ARTIFACTS.UPPER),
                                     p.adjust.method = "fdr", paired = TRUE, exact = FALSE, alternative = "greater")
    effect.sizes = get.effect.sizes(data.df, "artifact", w, p)

    ## put result in output file
    capture.output( print(w$p.value, print.gap = 3), file = stats.path.file, append = FALSE)
    cat("", file = stats.path.file, append = TRUE, fill = TRUE)
    capture.output( print(p$p.value, print.gap = 3), file = stats.path.file, append = TRUE)
    cat("", file = stats.path.file, append = TRUE, fill = TRUE)
    capture.output( print(effect.sizes, print.gap = 3), file = stats.path.file, append = TRUE)
}

##
## HYPOTHESIS 03
##

get.statistics.hypo4 = function(subset = NULL, null.model = FALSE) {

    logging::loginfo("Collecting the statistical test for hypothesis H4")

    ## set output path
    stats.path = file.path(TMP.STATS, "hypo4", if (null.model) NULL.STRING else "empirical")
    dir.create(stats.path, recursive = TRUE, showWarnings = FALSE)
    stats.path.file = file.path(stats.path, paste0("hypo4", if(!is.null(subset)) paste0("-", subset), ".txt"))

    ## create output file for Shapiro-Wilk tests
    shapiro.test.file = file.path(stats.path, paste0("hypo4", if(!is.null(subset)) paste0("-", subset), "-shapiro.txt"))
    cat("", file = shapiro.test.file, append = FALSE, fill = FALSE)

    data.df = lapply(CASESTUDIES, function(casestudy) {

        TMP.RANGES.FILLED = file.path(TMP, casestudy, "ranges", paste(casestudy, "ranges.filled", sep = "-"))
        load(TMP.RANGES.FILLED)

        ## get active ranges if wanted
        if (!is.null(subset) && subset == "active") {
            TMP.RANGES.ACTIVE = file.path(TMP, casestudy, "ranges", paste(casestudy, "ranges.active", sep = "-"))
            load(TMP.RANGES.ACTIVE)
            thresholds = attr(ranges.subset.active, "thresholds")
            ranges.subset.filled = join.boolean.values(list(
                ranges.subset.filled, ranges.subset.active
            ), op = "and")
        }

        results = collect.results(casestudy, ranges.subset.filled, null.model = null.model)
        artifacts = factor(results[["artifact"]], levels = ARTIFACTS)

        ## reduce results to range level (and remove all unnecessary stuff)
        cols = c("casestudy", "range", "artifact")
        cols.names = cols
        for (motif in MOTIFS) {
            cols = c(cols, paste0(motif, c(".collaborating", ".collaborating.and.mailing", ".p2")))
            cols.names = c(cols.names, paste0(motif, c(".cr", ".fcr", ".f")))
        }
        results = unique(results[cols])
        colnames(results) = cols.names

        ## construct melted data.frame
        results.melt = lapply(MOTIFS, function(motif) {
            ## melt data.frame
            cols = paste0(motif, c(".f")) # ".cr", ".fcr",
            melted = melt(results, id.vars = c("casestudy", "range", "artifact"), measure.vars = cols)
            ## add motif column
            melted = cbind(melted, motif = factor(motif, levels = MOTIFS))
            ## remove motif prefix from variable
            melted["variable"] = gsub(paste0(motif, "."), "", melted[["variable"]])

            ## run Shapiro-Wilk tests
            melted.split = split(melted, melted$artifact)
            for (melted.split.art in melted.split) {
                artifact.factor = unique(melted.split.art$artifact)
                artifact = levels(artifact.factor)[artifact.factor]
                ## run statistical test
                # browser()
                df.test.shapiro = shapiro.test(melted.split.art$value)
                ## put result in output file
                cat("####", casestudy, artifact, motif, file = shapiro.test.file, append = TRUE, fill = TRUE)
                capture.output( print(df.test.shapiro, print.gap = 3), file = shapiro.test.file, append = TRUE)
            }

            return(melted)
        })
        results.melt = rbind.fill(results.melt)

        return(results.melt)
    })
    data.df = rbind.fill(data.df)


    # ## 2) plots

    ## plots for this hypothesis are handled in the script 'history-plots' for now


    ## 3) statistics

    ## dump data to disk
    data.df.disk = dcast(data.df, "casestudy + range + motif ~ artifact")
    write.table(
        data.df.disk, file = file.path(stats.path, paste0("hypo4", if(!is.null(subset)) paste0("-", subset), ".csv")),
        sep = ";", row.names = FALSE
    )

    ## 3.1) chaos indicator

    # df.chaos = data.frame()
    # df.chaos.artifact = data.frame()
    # cat("", file = stats.path.file, append = FALSE, fill = FALSE)
    # for (casestudy in CASESTUDIES) {
    #
    #     df = data.df[data.df[["casestudy"]] == CASESTUDIES.REAL[casestudy], ]
    #
    #     for (motif in MOTIFS) {
    #         df.motif = df[df[["motif"]] == motif, ]
    #
    #         ## run statistical test
    #         w = pairwise.mann.whitney.test.w(df.motif[["value"]], factor(df.motif$artifact, levels = ARTIFACTS),
    #                                          p.adjust.method = "none", paired = TRUE, exact = FALSE, alternative = "greater")
    #         p = pairwise.mann.whitney.test.p(df.motif[["value"]], factor(df.motif$artifact, levels = ARTIFACTS),
    #                                          p.adjust.method = "fdr", paired = TRUE, exact = FALSE, alternative = "greater")
    #         effect.sizes = get.effect.sizes(df.motif, "artifact", w, p)
    #
    #         ## compute chaos indicator
    #         num.ranges = length(unique(df.motif[["range"]]))
    #         chaos.indicator.diffs = lapply(ARTIFACTS, function(artifact) {
    #             data.f = df.motif[df.motif[["artifact"]] == artifact, "value"]
    #             return(abs(diff(data.f, 1)))
    #         })
    #         chaos.indicator = sum(unlist(chaos.indicator.diffs), na.rm = TRUE) /
    #             (num.ranges * length(ARTIFACTS)) # mean by ranges and artifacts
    #         chaos.indicator.artifact = sapply(chaos.indicator.diffs, sum, na.rm = TRUE) / (num.ranges)
    #         df.chaos = rbind(
    #             df.chaos,
    #             data.frame(
    #                 "casestudy" = CASESTUDIES.REAL[[casestudy]],
    #                 "motif" = motif,
    #                 "num.ranges" = num.ranges,
    #                 "chaos.indicator" = chaos.indicator
    #             )
    #         )
    #         df.chaos.artifact = rbind(
    #             df.chaos.artifact,
    #             data.frame(
    #                 "casestudy" = CASESTUDIES.REAL[[casestudy]],
    #                 "motif" = motif,
    #                 "artifact" = ARTIFACTS,
    #                 "num.ranges" = num.ranges,
    #                 "chaos.indicator" = chaos.indicator.artifact
    #             )
    #         )
    #
    #         ## put result in output file
    #         sink(file = stats.path.file, append = TRUE, type = "output")
    #         cat("####", casestudy, motif, fill = TRUE)
    #         print(w$p.value, print.gap = 3)
    #         cat("", fill = TRUE)
    #         print(p$p.value, print.gap = 3)
    #         cat("", fill = TRUE)
    #         print(effect.sizes, print.gap = 3)
    #         cat("", fill = TRUE)
    #         cat("", fill = TRUE)
    #         sink()
    #     }
    #
    # }
    #
    # # dump chaos factor to disk
    # stats.chaos.file = file.path(stats.path, paste0("hypo4-chaos", if(!is.null(subset)) paste0("-", subset), ".csv"))
    # write.table(df.chaos, file = stats.chaos.file, sep = ";", row.names = FALSE)
    #
    # stats.chaos.artifact.file = file.path(stats.path, paste0("hypo4-chaos-by-artifact",
    #                                                          if(!is.null(subset)) paste0("-", subset), ".csv"))
    # df.chaos.artifact = dcast(df.chaos.artifact,
    #                           "casestudy + motif + num.ranges ~ artifact",
    #                           value.var = "chaos.indicator")
    # df.chaos.artifact = df.chaos.artifact[c( "casestudy", "motif", "num.ranges", ARTIFACTS)]
    # write.table(df.chaos.artifact, file = stats.chaos.artifact.file, sep = ";", row.names = FALSE)

    ## 3.2) fractal dimension

    ## configuration for fractal dimension
    FD.METHOD = "boxcount"
    FD.NLAGS = "all"

    ## calculation
    df.fracdim.artifact = data.frame()
    cat("", file = stats.path.file, append = FALSE, fill = FALSE)
    for (casestudy in CASESTUDIES) {

        df.casestudy = data.df[data.df[["casestudy"]] == CASESTUDIES.REAL[casestudy], ]

        for (artifact in ARTIFACTS.UPPER) {

            df.artifact = df.casestudy[df.casestudy[["artifact"]] == tolower(artifact), ]

            for (motif in MOTIFS.UPPER) {
                df.motif = df.artifact[df.artifact[["motif"]] == tolower(motif), ]
                values = df.motif[["value"]]

                ## remove all NA values to avoid time windows representing stagnation
                values = values[!is.na(values)]

                # ## identify all ranges at start that do have an NA value and set their value to zero
                # start = 0
                # for (i in seq_len(length(values))) {
                #     filled = !is.na(values[i])
                #     if (start == 0 && filled) start = i
                # }
                # if (start != 0) values[seq_len(start - 1)] = 0
                #
                # ## set all other NA values to previous value
                # for (idx in which(is.na(values))) {
                #     values[idx] = values[idx - 1]
                # }

                ## calculate fractal dimension using method FD.METHOD
                fd = fractaldim::fd.estimate(values, method = FD.METHOD, nlags = FD.NLAGS)$fd

                ## add data to results
                df.fracdim.artifact = rbind(
                    df.fracdim.artifact,
                    data.frame(
                        casestudy = CASESTUDIES.REAL[casestudy],
                        artifact = artifact,
                        motif = motif,
                        D = fd
                    )
                )
            }
        }
    }

    df.fracdim.disk = dcast(df.fracdim.artifact, "casestudy + artifact ~ motif", value.var = "D")
    write.table(
        df.fracdim.disk, file = file.path(stats.path, paste0("hypo4", if(!is.null(subset)) paste0("-", subset), ".txt")),
        sep = ";", row.names = FALSE
    )

    df.fracdim.disk.mean = dcast(df.fracdim.artifact, "casestudy ~ motif", value.var = "D",
                            fun.aggregate = mean, margin = "artifact")
    df.fracdim.disk.mean = df.fracdim.disk.mean[c(1, 3, 2)] # re-order columns
    df.fracdim.disk.mean = df.fracdim.disk.mean[order(df.fracdim.disk.mean$Triangle), ] # order rows by triangle motif
    write.table(
        df.fracdim.disk.mean, file = file.path(stats.path, paste0("hypo4", if(!is.null(subset)) paste0("-", subset),
                                                                  "-average-per-motif.txt")),
        sep = ";", row.names = FALSE
    )

    ## 4) print fractal dimension to LaTeX table

    tex.file = file.path(stats.path, paste0("hypo4", if(!is.null(subset)) paste0("-", subset), ".tex"))
    strCaption = paste0("Fractal-dimension values $D$ for all subject projects and motifs, sorted by $D_{\\motiftriangle}$ and grouped by similar values.\nThe groups of values are derived in combination with the plots in Figure~\\ref{fig:results:hypo4:triangle}.")

    ## properly set project names
    df.fracdim.disk.mean$casestudy = get.casestudy.print(df.fracdim.disk.mean$casestudy)

    ## construct separators:
    separators.pos = list()
    separators.command = c()
    ## 4.1.1) top rule + headings + mid rule
    separators.pos = c(separators.pos, list(-1))
    separators.command = c(separators.command,
                           paste(
                               "\\toprule\n",
                               paste(c("\\bfseries Project",
                                       "$\\mathbf{D_{\\motiftriangle}}$",
                                       "$\\mathbf{D_{\\motifsquare}}$ \\\\\n"),
                                     collapse = " & "),
                               "\\midrule \n"
                           ))
    ## 4.1.2) midrules for groups
    separators.pos.midrule = c(2, 5, 8)
    separators.pos = c(separators.pos, separators.pos.midrule)
    separators.command = c(separators.command, rep("\\addlinespace\n", times = length(separators.pos.midrule)))
    ## 4.1.3) bottom rule
    separators.pos = c(separators.pos, nrow(df.fracdim.disk.mean))
    separators.command = c(separators.command, paste("\\bottomrule\n"))
    ## 4.2) set up xtable output
    tab.tex = print(xtable(df.fracdim.disk.mean,
                           align = "llll", # align and put a vertical line (first "l" again represents column of row numbers)
                           caption = strCaption, label = "tab:results:fractal"),
                    file = tex.file,
                    floating.environment = "table",
                    table.placement = "bp",
                    # size = "normalsize", #Change size; useful for bigger tables "normalsize" "footnotesize"
                    include.rownames = FALSE, #Don't print rownames
                    include.colnames = FALSE, #We create them ourselves
                    caption.placement = "top", #"top", NULL
                    hline.after = NULL, #We don't need hline; we use booktabs
                    floating = TRUE, # whether \begin{Table} should be created (TRUE) or not (FALSE)
                    sanitize.text.function = force, # Important to treat content of first column as latex function
                    booktabs = TRUE,
                    add.to.row = list(
                        pos = separators.pos,
                        command = separators.command
                    )
    )
    ## 4.3) clean-up (fix alignments and new-line characters)
    tab.tex = gsub("begin\\{tabular\\}\\{lll\\}",
                   "sisetup{table-format=1.2}\n\\\\begin{tabular}{l SS}",
                   tab.tex)
    tab.tex = gsub("[[:space:]]*\\n", "\n", tab.tex)
    ## 4.4) output
    cat(tab.tex, file = tex.file)

}
