## libraries
library(igraph)
library(ggplot2)
library(plyr) # for match_df function
library(reshape2) # for melt function
library(gtools) # for combinations function
library(lubridate) # for date parsing and handling


construct.analysis.id = function(casestudy, artifact, artifact.relation) {
    analysis = paste(
        casestudy,
        artifact,
        artifact.relation,
        ## developer.relation,
        sep = "_"
    )
    return(analysis)
}


load.and.save = function(variable, dump.path, skip = FALSE, if.not.found = NA) {
    if (!skip && file.exists(dump.path)) {
        logging::logdebug("Load %s from previously dumped object: %s.", variable, dump.path)
        load(file = dump.path) # load the list named "variable" into the current environment
    } else {
        res = if.not.found()

        assign(variable, res) # rewrite to variable name
        rm(res) # clear memory

        logging::logdebug("Dumping object %s to %s.", variable, dump.path)
        save(list = variable, file = dump.path) # save automatically
    }

    return(get0(variable))
}


join.boolean.values = function(value.list, op = c("or", "and")) {
    op = match.arg(op)

    value.list.length = length(value.list)
    j = value.list[[1]]

    if (length(value.list) > 1) {
        for (vl in value.list[2:value.list.length]) {
            if (op == "or")
                j = j | vl
            else if (op == "and")
                j = j & vl
        }
        # browser(expr = op == "and")
    }

    return(j)
}

strip.attributes.from.network = function(net) {

    edge.attributes = igraph::list.edge.attributes(net)
    for (att in edge.attributes) {
        if (att %in% c())
            next
        net = igraph::remove.edge.attribute(net, att)
    }

    # vertex.attributes = igraph::list.vertex.attributes(net)
    # for (att in vertex.attributes) {
    #     if (att %in% c("name", "id"))
    #         next
    #     net = igraph::remove.vertex.attribute(net, att)
    # }

    return(net)
}

#############################

## TODO: this should be changed to use the Tanimoto coefficient, a more theoriticaly
##       sound and established similairty measure for weighted graphs
vertex.edge.weight.difference = function(g1, v1, g2, v2) {
    ##  -- To be used only on weighted graphs --
    ## computes the percent difference between two verteces from two different
    ## graphs. The percent difference is calculated based on the difference
    ## between the weights of common edges divided by the average edge weight
    ## -- Output --
    ## percent.difference: pertage DIFFERENCE for the matching edges

    g1.adjMat = get.adjacency(g1)
    g2.adjMat = get.adjacency(g2)

    in.v1  = neighbors(g1, v1, mode="in")
    out.v1 = neighbors(g1, v1, mode="out")
    in.v2  = neighbors(g2, v2, mode="in")
    out.v2 = neighbors(g2, v2, mode="out")

    in.union  = union(in.v1,in.v2)
    out.union = union(out.v1, out.v2)
    in.inter  = intersect(in.v1, in.v2)
    out.inter = intersect(out.v1,out.v2)

    if (is.weighted(g1) && is.weighted(g2)){

        out.percent.diff = 0
        in.percent.diff  = 0

        if (length(out.union) != 0){
            out.diff = abs(g1.adjMat[v1, out.union] - g2.adjMat[v2, out.union])
            out.avg = 0.5 * (g1.adjMat[v1, out.union] +
                                 g2.adjMat[v2, out.union])
            out.percent.diff = mean(out.diff / out.avg)
        }
        if (length(in.union) != 0){
            in.diff = abs(g1.adjMat[in.union, v1] - g2.adjMat[in.union, v2])
            in.avg  = 0.5 * (g1.adjMat[in.union, v1]  +
                                 g2.adjMat[in.union, v2])
            in.percent.diff  = mean(in.diff / in.avg)
        }
        percent.difference = mean(c(in.percent.diff, out.percent.diff))
    }
    else{
        e = simpleError("difference comparison not possible for unweighted
                        graphs")
        stop(e)
    }

    return(percent.difference)
}


vertex.neighborhood.difference = function(g1, v1, g2, v2) {
    ## Calculates the percent similarity using the Jaccard index concept from
    ## set theory. This considers the neighbour hoods of matching
    ## verteces from two different graphs and does not consider the edge weights
    ## rather only the existence of edges.
    ## -- Output --
    ## similarity: a percentage of how SIMILAR the neightboods are
    options(warn = -1)

    in.v1  = neighbors(g1, v1, mode = "in")
    out.v1 = neighbors(g1, v1, mode = "out")
    in.v2  = neighbors(g2, v2, mode = "in")
    out.v2 = neighbors(g2, v2, mode = "out")

    totalEdges = length(union(in.v1, in.v2)) + length(union(out.v1, out.v2))
    matchEdges = length(intersect(in.v1, in.v2)) + length(intersect(out.v1, out.v2))

    if (totalEdges != 0) {
        difference = 1 - (matchEdges / totalEdges)
    } else {
        difference = 0
    }

    options(warn = 0)

    return(difference)
}


graph.difference = function(g1, g2, weighted = FALSE) {
    ## two graphs on the same vertex set can be compared by considering
    ## the percent at which the two graphs agree on an edge

    ## compares to graphs that have match indexing, meaning that vertex 1 in g1
    ## is the same person in as vertex 1 in g2. If the graph is weighted then
    ## a difference is considered by calculating a percent difference on matching
    ## edges. If the graph is not weigthed then only the existance of edges
    ## is used to calculated how different the two graphs are.

    vertexList1 = V(g1)
    vertexList2 = V(g2)

    if (!all(vertexList1 == vertexList2)) {
        e = simpleError("graphs not compatible!")
        stop(e)
    } else {
        vertexList = vertexList1
    }

    vert.diff = numeric(length(vertexList))
    names(vert.diff) = vertexList1$name
    for (v in vertexList) {
        if (weighted) {
            vert.diff[v] = vertex.edge.weight.difference(g1, v, g2, v)
        } else {
            vert.diff[v] = vertex.neighborhood.difference(g1, v, g2, v)
        }
    }
    return(vert.diff)
}


## / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / /
## Date handling (taken from newer version of network library) -------------

#' Parse a date with optional time
#'
#' Notice: Time-zone suffixes are ignored. The \code{input} is expected to be in UTC,
#' even if the \code{input} contains another time-zone suffix.
#'
#' @param input The date string, a vector of date strings, or a list of date strings
#'
#' @return The parsed date(s) as POSIXct object, without changing the underlying data structure
get.date.from.string = function(input) {

    ## re-usable function to parse date strings with lubridate
    convert.text.to.date = function(text) {
        date = lubridate::ymd_hms(text, truncated = 3)
        return(date)
    }

    ## Handle list manually as lubridate would
    ## emit warnings on lists containing NA
    if (is.list(input)) {
        result = lapply(input, convert.text.to.date)
    } else {
        result = convert.text.to.date(input)
    }

    return(result)
}

#' Convert UNIX timestamp to POSIXct
#'
#' @param timestmap The timestamp
#'
#' @return The parsed date as POSIXct object
get.date.from.unix.timestamp = function(timestamp) {
    date = lubridate::as_datetime(timestamp)
    return(date)
}

#' Formats a given date as string using the format "%Y-%m-%d %H:%M:%S"
#'
#' @param input The POSIXct object, a vector of such, or a list of such
#'
#' @return The formatted date(s), without changing the underlying data structure
get.date.string = function(input) {

    ## re-usable function to parse date strings with lubridate
    convert.date.to.text = function(date) {

        ## if we do not have a POSIXct object here, do not convert
        if (!lubridate::is.POSIXct(date)) {
            return(date)
        }

        text = strftime(date, format = "%Y-%m-%d %H:%M:%S")
        return(text)
    }

    ## Handle list manually to not change the underlying
    ## data structure
    if (is.list(input)) {
        result = lapply(input, convert.date.to.text)
    } else {
        result = convert.date.to.text(input)
    }

    return(result)
}
